from __future__ import annotations

import itertools
import logging
import re
import sys
from typing import Optional
from urllib.parse import urlparse

import libsbml
import pandas as pd
from pydantic import BaseModel

from napistu import sbml_dfs_core
from napistu import utils

from napistu.constants import IDENTIFIERS
from napistu.constants import BIOLOGICAL_QUALIFIER_CODES
from napistu.constants import ENSEMBL_MOLECULE_TYPES_TO_ONTOLOGY
from napistu.constants import ENSEMBL_MOLECULE_TYPES_FROM_ONTOLOGY
from napistu.constants import ENSEMBL_SPECIES_FROM_CODE
from napistu.constants import ENSEMBL_SPECIES_TO_CODE
from napistu.constants import SPECIES_IDENTIFIERS_REQUIRED_VARS
from napistu.constants import SBML_DFS_SCHEMA
from napistu.constants import IDENTIFIERS_REQUIRED_VARS

logger = logging.getLogger(__name__)


class Identifiers:
    """
    Identifiers for a single entity or relationship.

    Attributes
    ----------
    ids : list
        a list of identifiers which are each a dict containing an ontology and identifier
    verbose : bool
        extra reporting, defaults to False

    Methods
    -------
    print
        Print a table of identifiers
    filter(ontologies, summarize)
        Returns a bool of whether 1+ of the ontologies was represented
    hoist(ontology)
        Returns value(s) from an ontology

    """

    def __init__(self, id_list: list, verbose: bool = False) -> None:
        """
        Tracks a set of identifiers and the ontologies they belong to.

        Parameters
        ----------
        id_list : list
            a list of identifier dictionaries containing ontology, identifier, and optionally url

        Returns
        -------
        None.

        """

        # read list and validate format
        validated_id_list = _IdentifiersValidator(id_list=id_list).model_dump()[
            "id_list"
        ]

        if (len(id_list) == 0) and verbose:
            logger.debug('zero identifiers in "id_list"')

        if len(id_list) != 0:
            # de-duplicate {identifier, ontology} tuples

            coded_ids = [
                x[IDENTIFIERS.ONTOLOGY] + "_" + x[IDENTIFIERS.IDENTIFIER]
                for x in validated_id_list
            ]
            unique_cids = []
            unique_cid_indices = []
            i = 0
            for cid in coded_ids:
                if cid not in unique_cids:
                    unique_cids.append(cid)
                    unique_cid_indices.append(i)
                i += 1
            validated_id_list = [validated_id_list[i] for i in unique_cid_indices]

        self.ids = validated_id_list

    def print(self):
        """Print a table of identifiers"""

        utils.style_df(pd.DataFrame(self.ids), hide_index=True)

    def filter(self, ontologies, summarize=True):
        """Returns a bool of whether 1+ of the ontologies was represented"""

        if isinstance(ontologies, str):
            ontologies = [ontologies]

        # filter based on whether any ontology of interest are present
        # identifier_matches = [x['ontology'] == y for x in self.ids for y in ontologies]

        identifier_matches = []
        for an_id in self.ids:
            identifier_matches.append(
                any([an_id[IDENTIFIERS.ONTOLOGY] == y for y in ontologies])
            )

        if summarize:
            return any(identifier_matches)
        else:
            return identifier_matches

    def hoist(self, ontology: str, squeeze: bool = True) -> str | list[str] | None:
        """Returns value(s) from an ontology

        Args:
            ontology (str): the ontology of interest
            squeeze (bool): if True, return a single value if possible

        Returns:
            str or list: the value(s) of an ontology of interest

        """

        if not isinstance(ontology, str):
            raise TypeError(f"{ontology} must be a str")

        # return the value(s) of an ontology of interest
        ontology_matches = [
            x for x, y in zip(self.ids, self.filter(ontology, summarize=False)) if y
        ]
        ontology_ids = [x[IDENTIFIERS.IDENTIFIER] for x in ontology_matches]

        if squeeze:
            if len(ontology_ids) == 0:
                return None
            elif len(ontology_ids) == 1:
                return ontology_ids[0]
        return ontology_ids


def merge_identifiers(identifier_series: pd.Series) -> Identifiers:
    """
    Aggregate Identifiers

    Merge a pd.Series of Identifiers objects into a single Identifiers object

    Args:
    identifier_series: pd.Series
        A pd.Series of of identifiers.Identifiers objects


    Returns:
    An identifiers.Identifiers object

    """

    if len(identifier_series) == 1:
        # if there is only a single entry then just return it because no merge is needed
        return identifier_series.iloc[0]
    else:
        # merge a list of identifiers objects into a single identifers object
        # Identifiers will remove redundancy
        merged_ids = list(
            itertools.chain.from_iterable(identifier_series.map(lambda x: x.ids))
        )
        return Identifiers(merged_ids)


def df_to_identifiers(df: pd.DataFrame, entity_type: str) -> pd.Series:
    """
    Convert a DataFrame of identifier information to a Series of Identifiers objects.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing identifier information with required columns:
        ontology, identifier, url, bqb
    index_col : str
        Name of the column to use as index for the output Series

    Returns
    -------
    pd.Series
        Series indexed by index_col containing Identifiers objects
    """

    if entity_type not in SBML_DFS_SCHEMA.SCHEMA:
        raise ValueError(f"Invalid entity type: {entity_type}")

    table_schema = SBML_DFS_SCHEMA.SCHEMA[entity_type]
    if "id" not in table_schema:
        raise ValueError(f"The entity type {entity_type} does not have an id column")

    table_pk_var = table_schema["pk"]
    expected_columns = set([table_pk_var]) | IDENTIFIERS_REQUIRED_VARS
    missing_columns = expected_columns - set(df.columns)
    if missing_columns:
        raise ValueError(
            f"The DataFrame does not contain the required columns: {missing_columns}"
        )

    # Process identifiers to remove duplicates
    indexed_df = (
        df
        # remove duplicated identifiers
        .groupby([table_pk_var, IDENTIFIERS.ONTOLOGY, IDENTIFIERS.IDENTIFIER])
        .first()
        .reset_index()
        .set_index(table_pk_var)
    )

    # create a dictionary of new Identifiers objects
    expanded_identifiers_dict = {
        i: _expand_identifiers_new_entries(i, indexed_df)
        for i in indexed_df.index.unique()
    }

    output = pd.Series(expanded_identifiers_dict).rename(table_schema["id"])
    output.index.name = table_pk_var

    return output


def format_uri(uri: str, biological_qualifier_type: str | None = None) -> Identifiers:
    """
    Convert a RDF URI into an Identifier object
    """

    identifier = format_uri_url(uri)

    if identifier is None:
        raise NotImplementedError(f"{uri} is not a valid way of specifying a uri")

    _validate_bqb(biological_qualifier_type)
    identifier[IDENTIFIERS.BQB] = biological_qualifier_type

    return identifier


def _validate_bqb(bqb):
    if bqb is None:
        logger.warning(
            '"biological_qualifier_type" is None; consider adding a valid '
            'BQB code. For a list of BQB codes see "BQB" in constants.py'
        )
    else:
        if not isinstance(bqb, str):
            raise TypeError(
                f"biological_qualifier_type was a {type(bqb)} and must be a str or None"
            )

        if not bqb.startswith("BQB"):
            raise ValueError(
                f"The provided BQB code was {bqb} and all BQB codes start with "
                'start with "BQB". Please either use a valid BQB code (see '
                '"BQB" in constansts.py) or use None'
            )


def format_uri_url(uri: str) -> dict:
    # check whether the uri is specified using a url
    result = urlparse(uri)
    if not all([result.scheme, result.netloc, result.path]):
        return None

    # valid url

    netloc = result.netloc
    split_path = result.path.split("/")

    try:
        if netloc == "identifiers.org":
            ontology, identifier = format_uri_url_identifiers_dot_org(split_path)
        elif netloc == "reactome.org":
            ontology = "reactome"
            identifier = split_path[-1]
        # genes and gene products
        elif netloc == "www.ensembl.org" and split_path[-1] == "geneview":
            ontology = "ensembl_gene"
            identifier, id_ontology, _ = parse_ensembl_id(result.query)  # type: ignore
            if ontology != id_ontology:
                raise ValueError(
                    f"Ontology mismatch: expected {ontology}, got {id_ontology}"
                )
        elif netloc == "www.ensembl.org" and split_path[-1] in [
            "transview",
            "Transcript",
        ]:
            ontology = "ensembl_transcript"
            identifier, id_ontology, _ = parse_ensembl_id(result.query)  # type: ignore
            if ontology != id_ontology:
                raise ValueError(
                    f"Ontology mismatch: expected {ontology}, got {id_ontology}"
                )
        elif netloc == "www.ensembl.org" and split_path[-1] == "ProteinSummary":
            ontology = "ensembl_protein"
            identifier, id_ontology, _ = parse_ensembl_id(result.query)  # type: ignore
            if ontology != id_ontology:
                raise ValueError(
                    f"Ontology mismatch: expected {ontology}, got {id_ontology}"
                )
        elif netloc == "www.ensembl.org" and (
            re.search("ENS[GTP]", split_path[-1])
            or re.search("ENS[A-Z]{3}[GTP]", split_path[-1])
        ):
            # format ensembl IDs which lack gene/transview
            identifier, ontology, _ = parse_ensembl_id(split_path[-1])

        elif netloc == "www.mirbase.org" or netloc == "mirbase.org":
            ontology = "mirbase"
            if re.search("MI[0-9]+", split_path[-1]):
                identifier = utils.extract_regex_search("MI[0-9]+", split_path[-1])
            elif re.search("MIMAT[0-9]+", split_path[-1]):
                identifier = utils.extract_regex_search("MIMAT[0-9]+", split_path[-1])
            elif re.search("MI[0-9]+", result.query):
                identifier = utils.extract_regex_search("MI[0-9]+", result.query)
            elif re.search("MIMAT[0-9]+", result.query):
                identifier = utils.extract_regex_search("MIMAT[0-9]+", result.query)
            else:
                raise TypeError(
                    f"{result.query} does not appear to match MiRBase identifiers"
                )
        elif netloc == "purl.uniprot.org":
            ontology = "uniprot"
            identifier = split_path[-1]
        elif netloc == "rnacentral.org":
            ontology = "rnacentral"
            identifier = split_path[-1]
        # chemicals
        elif split_path[1] == "chebi":
            ontology = "chebi"
            identifier = utils.extract_regex_search("[0-9]+$", result.query)
        elif netloc == "pubchem.ncbi.nlm.nih.gov":
            ontology = "pubchem"
            if result.query != "":
                identifier = utils.extract_regex_search("[0-9]+$", result.query)
            else:
                identifier = utils.extract_regex_search("[0-9]+$", split_path[-1])
        elif netloc == "www.genome.ad.jp":
            ontology = "genome_net"
            identifier = utils.extract_regex_search("[A-Za-z]+:[0-9]+$", uri)
        elif (
            netloc == "www.guidetopharmacology.org"
            and split_path[-1] == "LigandDisplayForward"
        ):
            ontology = "grac"
            identifier = utils.extract_regex_search("[0-9]+$", result.query)
        elif netloc == "www.chemspider.com" or netloc == "chemspider.com":
            ontology = "chemspider"
            identifier = split_path[-1]
        # reactions
        elif split_path[1] == "ec-code":
            ontology = "ec-code"
            identifier = split_path[-1]
        elif netloc == "www.rhea-db.org":
            ontology = "rhea"
            identifier = utils.extract_regex_search("[0-9]+$", result.query)
        # misc
        elif split_path[1] == "ols":
            ontology = "ols"
            identifier = split_path[-1]
        elif split_path[1] == "QuickGO":
            ontology = "go"
            identifier = split_path[-1]
        elif split_path[1] == "pubmed":
            ontology = "pubmed"
            identifier = split_path[-1]
        # DNA sequences
        elif netloc == "www.ncbi.nlm.nih.gov" and split_path[1] == "nuccore":
            ontology = "ncbi_refseq"
            identifier = split_path[-1]
        elif netloc == "www.ncbi.nlm.nih.gov" and split_path[1] == "sites":
            ontology = "ncbi_entrez_" + utils.extract_regex_search(
                "db=([A-Za-z0-9]+)\\&", result.query, 1
            )
            identifier = utils.extract_regex_search(
                r"term=([A-Za-z0-9\-]+)$", result.query, 1
            )
        elif netloc == "www.ebi.ac.uk" and split_path[1] == "ena":
            ontology = "ebi_refseq"
            identifier = split_path[-1]
        elif netloc == "www.thesgc.org" and split_path[1] == "structures":
            ontology = "sgc"
            identifier = split_path[-2]
        elif netloc == "www.mdpi.com":
            ontology = "mdpi"
            identifier = "/".join([i for i in split_path[1:] if i != ""])
        elif netloc == "dx.doi.org":
            ontology = "dx_doi"
            identifier = "/".join(split_path[1:])
        elif netloc == "doi.org":
            ontology = "doi"
            identifier = "/".join(split_path[1:])
        elif netloc == "www.ncbi.nlm.nih.gov" and split_path[1] == "books":
            ontology = "ncbi_books"
            identifier = split_path[2]
        elif netloc == "www.ncbi.nlm.nih.gov" and split_path[1] == "gene":
            ontology = "ncbi_gene"
            identifier = split_path[2]
        elif netloc == "www.phosphosite.org":
            ontology = "phosphosite"
            identifier = utils.extract_regex_match(".*id=([0-9]+).*", uri)
        elif netloc == "ncithesaurus.nci.nih.gov":
            ontology = "NCI_Thesaurus"
            identifier = utils.extract_regex_match(".*code=([0-9A-Z]+).*", uri)
        elif netloc == "matrixdb.ibcp.fr":
            molecule_class = utils.extract_regex_match(
                ".*class=([a-zA-Z]+).*", uri
            ).lower()
            ontology = f"matrixdb_{molecule_class}"
            identifier = utils.extract_regex_match(".*name=([0-9A-Za-z]+).*", uri)
        elif netloc == "matrixdb.univ-lyon1.fr":
            molecule_class = utils.extract_regex_match(
                ".*type=([a-zA-Z]+).*", uri
            ).lower()
            ontology = f"matrixdb_{molecule_class}"
            identifier = utils.extract_regex_match(".*value=([0-9A-Za-z]+).*", uri)
        else:
            raise NotImplementedError(
                f"{netloc} in the {uri} url has not been associated with a known ontology"
            )
    except TypeError:
        logger.warning(
            f"An identifier could not be found using the specified regex for {uri} based on the {ontology} ontology"
        )
        logger.warning(result)
        logger.warning("ERROR")
        sys.exit(1)

    # rename some entries

    if ontology == "ncbi_gene":
        ontology = "ncbi_entrez_gene"

    id_dict = {"ontology": ontology, "identifier": identifier, "url": uri}

    return id_dict


def parse_ensembl_id(input_str: str) -> tuple[str, str, str]:
    """
    Parse Ensembl ID

    Extract the molecule type and species name from a string containing an ensembl identifier.

    Args:
        input_str (str):
            A string containing an ensembl gene, transcript, or protein identifier

    Returns:
        identifier (str):
            The substring matching the full identifier
        molecule_type (str):
            The ontology the identifier belongs to:
                - G -> ensembl_gene
                - T -> ensembl_transcript
                - P -> ensembl_protein
        species (str):
            The species name the identifier belongs to

    """

    # validate that input is an ensembl ID
    if not re.search("ENS[GTP][0-9]+", input_str) and not re.search(
        "ENS[A-Z]{3}[GTP][0-9]+", input_str
    ):
        ValueError(
            f"{input_str} did not match the expected formats of an ensembl identifier:",
            "ENS[GTP][0-9]+ or ENS[A-Z]{3}[GTP][0-9]+",
        )

    # extract the species code (three letters after ENS if non-human)
    species_code_search = re.compile("ENS([A-Z]{3})?[GTP]").search(input_str)

    if species_code_search.group(1) is None:  # type:ignore
        species = "Homo sapiens"
        molecule_type_regex = "ENS([GTP])"
        id_regex = "ENS[GTP][0-9]+"
    else:
        species_code = species_code_search.group(1)  # type:ignore

        if species_code not in ENSEMBL_SPECIES_FROM_CODE.keys():
            raise ValueError(
                f"The species code for {input_str}: {species_code} did not "
                "match any of the entries in ENSEMBL_SPECIES_CODE_LOOKUPS."
            )

        species = ENSEMBL_SPECIES_FROM_CODE[species_code]
        molecule_type_regex = "ENS[A-Z]{3}([GTP])"
        id_regex = "ENS[A-Z]{3}[GTP][0-9]+"

    # extract the molecule type (genes, transcripts or proteins)
    molecule_type_code_search = re.compile(molecule_type_regex).search(input_str)
    if not molecule_type_code_search:
        raise ValueError(
            "The ensembl molecule code (i.e., G, T or P) could not be extracted from {input_str}"
        )
    else:
        molecule_type_code = molecule_type_code_search.group(1)  # type: str

    if molecule_type_code not in ENSEMBL_MOLECULE_TYPES_TO_ONTOLOGY.keys():
        raise ValueError(
            f"The molecule type code for {input_str}: {molecule_type_code} did not "
            "match ensembl genes (G), transcripts (T), or proteins (P)."
        )

    molecule_type = ENSEMBL_MOLECULE_TYPES_TO_ONTOLOGY[molecule_type_code]  # type: str

    identifier = utils.extract_regex_search(id_regex, input_str)  # type: str

    return identifier, molecule_type, species


def format_uri_url_identifiers_dot_org(split_path: list[str]):
    """Parse identifiers.org identifiers

    The identifiers.org identifier have two different formats:
    1. http://identifiers.org/<ontology>/<id>
    2. http://identifiers.org/<ontology>:<id>

    Currently we are identifying the newer format 2. by
    looking for the `:` in the second element of the split path.

    Also the ontology is converted to lower case letters.

    Args:
        split_path (list[str]): split url path

    Returns:
        tuple[str, str]: ontology, identifier
    """

    # formatting for the identifiers.org meta ontology

    # meta ontologies

    # identify old versions without `:`
    V2_SEPARATOR = ":"
    if V2_SEPARATOR in split_path[1]:
        # identifiers.org switched to format <ontology>:<id>
        path = "/".join(split_path[1:])
        if path.count(V2_SEPARATOR) != 1:
            raise ValueError(
                "The assumption is that there is only one ':'"
                f"in an identifiers.org url. Found more in: {path}"
            )
        ontology, identifier = path.split(":")
        ontology = ontology.lower()
    else:
        ontology = split_path[1]

        if ontology in ["chebi"]:
            identifier = utils.extract_regex_search("[0-9]+$", split_path[-1])
        elif len(split_path) != 3:
            identifier = "/".join(split_path[2:])
        else:
            identifier = split_path[-1]

    return ontology, identifier


def cv_to_Identifiers(entity):
    """
    Convert an SBML controlled vocabulary element into a cpr Identifiers object.

    Parameters:
    entity: libsbml.Species
        An entity (species, reaction, compartment, ...) with attached CV terms

    Returns:


    """

    # TO DO: add qualifier type http://sbml.org/Software/libSBML/5.18.0/docs/python-api/classlibsbml_1_1_c_v_term.html#a6a613cc17c6f853cf1c68da59286b373

    cv_list = list()
    for cv in entity.getCVTerms():
        if cv.getQualifierType() != libsbml.BIOLOGICAL_QUALIFIER:
            # only care about biological annotations
            continue

        biological_qualifier_type = BIOLOGICAL_QUALIFIER_CODES[
            cv.getBiologicalQualifierType()
        ]
        out_list = list()
        for i in range(cv.getNumResources()):
            try:
                out_list.append(
                    format_uri(cv.getResourceURI(i), biological_qualifier_type)
                )
            except NotImplementedError:
                logger.warning("Not all identifiers resolved: ", exc_info=True)

        cv_list.extend(out_list)
    return Identifiers(cv_list)


def create_uri_url(ontology: str, identifier: str, strict: bool = True) -> str:
    """
    Create URI URL

    Convert from an identifier and ontology to a URL reference for the identifier

    Parameters:
    ontology (str): An ontology for organizing genes, metabolites, etc.
    identifier (str): A systematic identifier from the \"ontology\" ontology.
    strict (bool): if strict then throw errors for invalid IDs otherwise return None

    Returns:
    url (str): A url representing a unique identifier

    """

    # default to no id_regex
    id_regex = None

    if ontology in ["ensembl_gene", "ensembl_transcript", "ensembl_protein"]:
        id_regex, url = ensembl_id_to_url_regex(identifier, ontology)
    elif ontology == "bigg.metabolite":
        url = f"http://identifiers.org/bigg.metabolite/{identifier}"
    elif ontology == "chebi":
        id_regex = "^[0-9]+$"
        url = f"http://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:{identifier}"
    elif ontology == "ec-code":
        id_regex = "^[0-9]+\\.[0-9]+\\.[0-9]+(\\.[0-9]+)?$"
        url = f"https://identifiers.org/ec-code/{identifier}"
    elif ontology == "envipath":
        url = f"http://identifiers.org/envipath/{identifier}"
    elif ontology == "go":
        id_regex = "^GO:[0-9]{7}$"
        url = f"https://www.ebi.ac.uk/QuickGO/term/{identifier}"
    elif ontology == "ncbi_entrez_gene":
        url = f"https://www.ncbi.nlm.nih.gov/gene/{identifier}"
    elif ontology == "ncbi_entrez_pccompound":
        id_regex = "^[A-Z]{14}\\-[A-Z]{10}\\-[A-Z]{1}$"
        url = f"http://www.ncbi.nlm.nih.gov/sites/entrez?cmd=search&db=pccompound&term={identifier}"
    elif ontology == "pubchem":
        id_regex = "^[0-9]+$"
        url = f"http://pubchem.ncbi.nlm.nih.gov/compound/{identifier}"
    elif ontology == "pubmed":
        id_regex = "^[0-9]+$"
        url = f"http://www.ncbi.nlm.nih.gov/pubmed/{identifier}"
    elif ontology == "reactome":
        id_regex = "^R\\-[A-Z]{3}\\-[0-9]{7}$"
        url = f"https://reactome.org/content/detail/{identifier}"
    elif ontology == "uniprot":
        id_regex = "^[A-Z0-9]+$"
        url = f"https://purl.uniprot.org/uniprot/{identifier}"
    elif ontology == "sgc":
        id_regex = "^[0-9A-Z]+$"
        url = f"https://www.thesgc.org/structures/structure_description/{identifier}/"
    elif ontology == "mdpi":
        id_regex = None
        url = f"https://www.mdpi.com/{identifier}"
    elif ontology == "mirbase":
        id_regex = None
        if re.match("MIMAT[0-9]", identifier):
            url = f"https://www.mirbase.org/mature/{identifier}"
        elif re.match("MI[0-9]", identifier):
            url = f"https://www.mirbase.org/hairpin/{identifier}"
        else:
            raise NotImplementedError(f"url not defined for this MiRBase {identifier}")
    elif ontology == "rnacentral":
        id_regex = None
        url = f"https://rnacentral.org/rna/{identifier}"
    elif ontology == "chemspider":
        id_regex = "^[0-9]+$"
        url = f"https://www.chemspider.com/{identifier}"

    elif ontology == "dx_doi":
        id_regex = r"^[0-9]+\.[0-9]+$"
        url = f"https://dx.doi.org/{identifier}"
    elif ontology == "doi":
        id_regex = None
        url = f"https://doi.org/{identifier}"

    elif ontology == "ncbi_books":
        id_regex = "^[0-9A-Z]+$"
        url = f"http://www.ncbi.nlm.nih.gov/books/{identifier}/"

    elif ontology == "ncbi_entrez_gene":
        id_regex = "^[0-9]+$"
        url = f"https://www.ncbi.nlm.nih.gov/gene/{identifier}"
    elif ontology == "phosphosite":
        id_regex = "^[0-9]+$"
        url = f"https://www.phosphosite.org/siteAction.action?id={identifier}"
    elif ontology == "NCI_Thesaurus":
        id_regex = "^[A-Z][0-9]+$"
        url = f"https://ncithesaurus.nci.nih.gov/ncitbrowser/ConceptReport.jsp?dictionary=NCI_Thesaurus&code={identifier}"
    elif ontology == "matrixdb_biomolecule":
        id_regex = "^[0-9A-Za-z]+$"
        url = f"http://matrixdb.univ-lyon1.fr/cgi-bin/current/newPort?type=biomolecule&value={identifier}"
    else:
        raise NotImplementedError(
            f"No identifier -> url logic exists for the {ontology} ontology in create_uri_url()"
        )

    # validate identifier with regex if one exists
    if id_regex is not None:
        if re.search(id_regex, identifier) is None:
            failure_msg = f"{identifier} is not a valid {ontology} id, it did not match the regex: {id_regex}"
            if strict:
                raise TypeError(failure_msg)
            else:
                print(failure_msg + " returning None")
                return None

    return url


def ensembl_id_to_url_regex(identifier: str, ontology: str) -> tuple[str, str]:
    """
    Ensembl ID to URL and Regex

    Map an ensembl ID to a validation regex and its canonical url on ensembl

    Args:
        identifier: str
            A standard identifier from ensembl genes, transcripts, or proteins
        ontology: str
            The standard ontology (ensembl_gene, ensembl_transcript, or ensembl_protein)

    Returns:
        id_regex: a regex which should match a valid entry in this ontology
        url: the id's url on ensembl
    """

    # extract the species name from the 3 letter species code in the id
    # (these letters are not present for humans)
    identifier, implied_ontology, species = parse_ensembl_id(identifier)  # type: ignore
    if implied_ontology != ontology:
        raise ValueError(
            f"Implied ontology mismatch: expected {ontology}, got {implied_ontology}"
        )

    # create an appropriate regex for validating input
    # this provides testing for other identifiers even if it is redundant with other
    # validation of ensembl ids

    if species == "Homo sapiens":
        species_code = ""
    else:
        species_code = ENSEMBL_SPECIES_TO_CODE[species]
    molecule_type_code = ENSEMBL_MOLECULE_TYPES_FROM_ONTOLOGY[ontology]

    id_regex = "ENS" + species_code + molecule_type_code + "[0-9]{11}"

    # convert to species format in ensembl urls
    species_url_field = re.sub(" ", "_", species)

    if ontology == "ensembl_gene":
        url = f"http://www.ensembl.org/{species_url_field}/geneview?gene={identifier}"
    elif ontology == "ensembl_transcript":
        url = f"http://www.ensembl.org/{species_url_field}/Transcript?t={identifier}"
    elif ontology == "ensembl_protein":
        url = f"https://www.ensembl.org/{species_url_field}/Transcript/ProteinSummary?t={identifier}"
    else:
        ValueError(f"{ontology} not defined")

    return id_regex, url


def check_reactome_identifier_compatibility(
    reactome_series_a: pd.Series,
    reactome_series_b: pd.Series,
) -> None:
    """
    Check Reactome Identifier Compatibility

    Determine whether two sets of Reactome identifiers are from the same species.

    Args:
        reactome_series_a: pd.Series
            a Series containing Reactome identifiers
        reactome_series_b: pd.Series
            a Series containing Reactome identifiers

    Returns:
        None

    """

    a_species, a_species_counts = _infer_primary_reactome_species(reactome_series_a)
    b_species, b_species_counts = _infer_primary_reactome_species(reactome_series_b)

    if a_species != b_species:
        a_name = reactome_series_a.name
        if a_name is None:
            a_name = "unnamed"

        b_name = reactome_series_b.name
        if b_name is None:
            b_name = "unnamed"

        raise ValueError(
            "The two provided pd.Series containing Reactome identifiers appear to be from different species. "
            f"The pd.Series named {a_name} appears to be {a_species} with {a_species_counts} examples of this code. "
            f"The pd.Series named {b_name} appears to be {b_species} with {b_species_counts} examples of this code."
        )

    return None


def _infer_primary_reactome_species(reactome_series: pd.Series) -> tuple[str, int]:
    """Infer the best supported species based on a set of Reactome identifiers"""

    series_counts = _count_reactome_species(reactome_series)

    if "ALL" in series_counts.index:
        series_counts = series_counts.drop("ALL", axis=0)

    return series_counts.index[0], series_counts.iloc[0]


def _count_reactome_species(reactome_series: pd.Series) -> pd.Series:
    """Count the number of species tags in a set of reactome IDs"""

    return (
        reactome_series.drop_duplicates().transform(_reactome_id_species).value_counts()
    )


def _reactome_id_species(reactome_id: str) -> str:
    """Extract the species code from a Reactome ID"""

    reactome_match = re.match("^R\\-([A-Z]{3})\\-[0-9]+", reactome_id)
    if reactome_match:
        try:
            value = reactome_match[1]
        except ValueError:
            raise ValueError(f"{reactome_id} is not a valid reactome ID")
    else:
        raise ValueError(f"{reactome_id} is not a valid reactome ID")

    return value


def _format_Identifiers_pubmed(pubmed_id: str) -> Identifiers:
    """
    Format Identifiers for a single PubMed ID.

    These will generally be used in an r_Identifiers field.
    """

    # create a url for lookup and validate the pubmed id
    url = create_uri_url(ontology="pubmed", identifier=pubmed_id, strict=False)
    id_entry = format_uri(uri=url, biological_qualifier_type="BQB_IS_DESCRIBED_BY")

    return Identifiers([id_entry])


def _check_species_identifiers_table(
    species_identifiers: pd.DataFrame,
    required_vars: set = SPECIES_IDENTIFIERS_REQUIRED_VARS,
):
    missing_required_vars = required_vars.difference(
        set(species_identifiers.columns.tolist())
    )
    if len(missing_required_vars) > 0:
        raise ValueError(
            f"{len(missing_required_vars)} required variables "
            "were missing from the species_identifiers table: "
            f"{', '.join(missing_required_vars)}"
        )

    return None


def _prepare_species_identifiers(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    dogmatic: bool = False,
    species_identifiers: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    """Accepts and validates species_identifiers, or extracts a fresh table if None."""

    if species_identifiers is None:
        species_identifiers = sbml_dfs.get_characteristic_species_ids(dogmatic=dogmatic)
    else:
        # check for compatibility
        try:
            # check species_identifiers format

            _check_species_identifiers_table(species_identifiers)
            # quick check for compatibility between sbml_dfs and species_identifiers
            _validate_assets_sbml_ids(sbml_dfs, species_identifiers)
        except ValueError as e:
            logger.warning(
                f"The provided identifiers are not compatible with your `sbml_dfs` object. Extracting a fresh species identifier table. {e}"
            )
            species_identifiers = sbml_dfs.get_characteristic_species_ids(
                dogmatic=dogmatic
            )

    return species_identifiers


def _validate_assets_sbml_ids(
    sbml_dfs: sbml_dfs_core.SBML_dfs, identifiers_df: pd.DataFrame
) -> None:
    """Check an sbml_dfs file and identifiers table for inconsistencies."""

    joined_species_w_ids = sbml_dfs.species.merge(
        identifiers_df[["s_id", "s_name"]].drop_duplicates(),
        left_index=True,
        right_on="s_id",
    )

    inconsistent_names_df = joined_species_w_ids.query("s_name_x != s_name_y").dropna()
    inconsistent_names_list = [
        f"{x} != {y}"
        for x, y in zip(
            inconsistent_names_df["s_name_x"], inconsistent_names_df["s_name_y"]
        )
    ]

    if len(inconsistent_names_list):
        example_inconsistent_names = inconsistent_names_list[
            0 : min(10, len(inconsistent_names_list))
        ]

        raise ValueError(
            f"{len(inconsistent_names_list)} species names do not match between "
            f"sbml_dfs and identifiers_df including: {', '.join(example_inconsistent_names)}"
        )

    return None


def _expand_identifiers_new_entries(
    sysid: str, expanded_identifiers_df: pd.DataFrame
) -> Identifiers:
    """Create an identifiers object from an index entry in a dataframe"""
    entry = expanded_identifiers_df.loc[sysid]

    if type(entry) is pd.Series:
        sysis_id_list = [entry.to_dict()]
    else:
        # multiple annotations
        sysis_id_list = list(entry.reset_index(drop=True).T.to_dict().values())

    return Identifiers(sysis_id_list)


class _IdentifierValidator(BaseModel):
    ontology: str
    identifier: str
    url: Optional[str] = None
    bqb: Optional[str] = None


class _IdentifiersValidator(BaseModel):
    id_list: list[_IdentifierValidator]



import copy
import logging
from typing import Optional, Union, Set, Dict

import pandas as pd

from napistu.constants import SBML_DFS, ONTOLOGIES_LIST
from napistu.matching.constants import (
    FEATURE_ID_VAR_DEFAULT,
    RESOLVE_MATCHES_AGGREGATORS,
    RESOLVE_MATCHES_TMP_WEIGHT_COL,
    BIND_DICT_OF_WIDE_RESULTS_STRATEGIES,
    BIND_DICT_OF_WIDE_RESULTS_STRATEGIES_LIST,
)
from napistu import identifiers, utils
from napistu.matching.species import match_features_to_wide_pathway_species
from napistu import sbml_dfs_core

logger = logging.getLogger(__name__)


def bind_wide_results(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    results_df: pd.DataFrame,
    results_name: str,
    ontologies: Optional[Union[Set[str], Dict[str, str]]] = None,
    dogmatic: bool = False,
    species_identifiers: Optional[pd.DataFrame] = None,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
    numeric_agg: str = RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN,
    keep_id_col: bool = True,
    verbose: bool = False,
    inplace: bool = True,
) -> Optional[sbml_dfs_core.SBML_dfs]:
    """
    Binds wide results to a sbml_dfs object.

    Take a table with molecular species-level attributes tied to systematic identifiers and match them to an sbml_dfs_model transferring these attributes to species_data

    Parameters
    ----------
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The sbml_dfs object to bind the results to.
    results_df : pd.DataFrame
        The table containing the results to bind.
    results_name : str
        The name of the results to bind.
    ontologies : Optional[Union[Set[str], Dict[str, str]]], default=None
        Either:
        - Set of columns to treat as ontologies (these should be entries in ONTOLOGIES_LIST )
        - Dict mapping wide column names to ontology names in the ONTOLOGIES_LIST controlled vocabulary
        - None to automatically detect valid ontology columns based on ONTOLOGIES_LIST
    dogmatic : bool
        Whether to respect differences between genes, transcripts, and proteins (True) or ignore them (False).
    species_identifiers : Optional[pd.DataFrame]
        Systematic identifiers for the molecular species "sbml_dfs". If None this will be generate on-the-fly.
    feature_id_var : str
        The name of the column in the results_df that contains the feature identifiers. If this does not exist it will be created.
    numeric_agg : str
        The aggregation method to use for resolving degeneracy.
    keep_id_col : bool
        Whether to keep the identifier column in the results_df.
    verbose : bool
        Whether to log cases of 1-to-many and many-to-one mapping and to indicate the behavior for resolving degeneracy
    inplace : bool, default=True
        Whether to modify the sbml_dfs object in place. If False, returns a copy.

    Returns
    -------
    sbml_dfs : sbml_dfs_core.SBML_dfs
        The sbml_dfs object with the results bound.
    """

    if not inplace:
        sbml_dfs = copy.deepcopy(sbml_dfs)

    species_identifiers = identifiers._prepare_species_identifiers(
        sbml_dfs, dogmatic=dogmatic, species_identifiers=species_identifiers
    )

    # match
    matched_s_ids_from_wide = match_features_to_wide_pathway_species(
        results_df,
        species_identifiers,
        ontologies=ontologies,
        feature_id_var=feature_id_var,
        verbose=verbose,
    )

    disambiguated_matches = resolve_matches(
        matched_data=matched_s_ids_from_wide,
        feature_id_var=feature_id_var,
        numeric_agg=numeric_agg,
        keep_id_col=keep_id_col,
    )

    clean_species_data = utils.drop_extra_cols(
        results_df, disambiguated_matches, always_include=[feature_id_var]
    )

    sbml_dfs.add_species_data(results_name, clean_species_data)

    return None if inplace else sbml_dfs


def bind_dict_of_wide_results(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    results_dict: dict,
    results_name: str,
    strategy: str = BIND_DICT_OF_WIDE_RESULTS_STRATEGIES.CONTATENATE,
    species_identifiers: pd.DataFrame = None,
    ontologies: Optional[Union[str, list]] = None,
    dogmatic: bool = False,
    inplace: bool = True,
    verbose=True,
):
    """
    Bind a dictionary of wide results to an SBML_dfs object.

    This function is used to bind a dictionary of wide results to 1 or more species_data attributes of an SBML_dfs object.
    The dictionary should have keys which are the modality names and values which are the results dataframes.
    The "strategy" argument controls how the results are added to the SBML_dfs object.

    Parameters
    ----------
    sbml_dfs : SBML_dfs
        The SBML_dfs object to bind the results to.
    results_dict : dict
        A dictionary of results dataframes with modality names as keys.
    results_name : str
        The name of the species_data attribute to bind the results to.
    strategy : str
        The strategy to use for binding the results.

        Options are:
        - "concatenate" : concatenate the results dataframes and add them as a single attribute.
        - "multiple_keys" : add each modality's results as a separate attribute. The attribute name will be f'{results_name}_{modality}'.
        - "stagger" : add each modality's results as a separate attribute. The attribute name will be f'{attr_name}_{modality}'.

    species_identifiers : pd.DataFrame
        A dataframe with species identifiers.
    ontologies : optional str, list
        The ontology to use for the species identifiers. If not provided, the column names of the results dataframes which match ONTOLOGIES_LIST will be used.
    dogmatic : bool
        Whether to use dogmatic mode. Ignored if species_identifiers is provided.
    verbose : bool
        Whether to print verbose output.
    inplace : bool, default=True
        Whether to modify the sbml_dfs object in place. If False, returns a copy.

    Returns
    -------
    Optional[SBML_dfs]
        If inplace=True, returns None. Otherwise returns the modified copy of sbml_dfs.
    """

    # validate strategy
    if strategy not in BIND_DICT_OF_WIDE_RESULTS_STRATEGIES_LIST:
        raise ValueError(
            f"Invalid strategy: {strategy}. Must be one of {BIND_DICT_OF_WIDE_RESULTS_STRATEGIES_LIST}"
        )

    species_identifiers = identifiers._prepare_species_identifiers(
        sbml_dfs, dogmatic=dogmatic, species_identifiers=species_identifiers
    )

    if not inplace:
        sbml_dfs = copy.deepcopy(sbml_dfs)

    if strategy == BIND_DICT_OF_WIDE_RESULTS_STRATEGIES.MULTIPLE_KEYS:
        for modality, results_df in results_dict.items():
            valid_ontologies = _get_wide_results_valid_ontologies(
                results_df, ontologies
            )

            modality_results_name = f"{results_name}_{modality}"

            bind_wide_results(
                sbml_dfs,
                results_df,
                modality_results_name,
                species_identifiers=species_identifiers,
                ontologies=valid_ontologies,
                inplace=True,  # Always use inplace=True here since we handle copying above
                verbose=verbose,
            )

        return None if inplace else sbml_dfs

    # create either a concatenated or staggered results table
    if strategy == BIND_DICT_OF_WIDE_RESULTS_STRATEGIES.CONTATENATE:
        results_df = pd.concat(results_dict.values(), axis=0)
    elif strategy == BIND_DICT_OF_WIDE_RESULTS_STRATEGIES.STAGGER:

        results_dict_copy = results_dict.copy()
        for k, v in results_dict_copy.items():
            valid_ontologies = _get_wide_results_valid_ontologies(v, ontologies)

            if verbose:
                logger.info(
                    f"Modality {k} has ontologies {valid_ontologies}. Other variables will be renamed to {k}_<variable>"
                )

            # rename all the columns besides ontologies names
            for var in v.columns:
                if var not in valid_ontologies:
                    results_dict_copy[k].rename(
                        columns={var: f"{var}_{k}"}, inplace=True
                    )

        results_df = pd.concat(results_dict_copy.values(), axis=1)

    valid_ontologies = _get_wide_results_valid_ontologies(results_df, ontologies)

    bind_wide_results(
        sbml_dfs,
        results_df,
        results_name,
        species_identifiers=species_identifiers,
        ontologies=valid_ontologies,
        inplace=True,  # Always use inplace=True here since we handle copying above
        verbose=verbose,
    )

    return None if inplace else sbml_dfs


def resolve_matches(
    matched_data: pd.DataFrame,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
    index_col: str = SBML_DFS.S_ID,
    numeric_agg: str = RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN,
    keep_id_col: bool = True,
) -> pd.DataFrame:
    """
    Resolve many-to-1 and 1-to-many matches in matched data.

    Parameters
    ----------
    matched_data : pd.DataFrame
        DataFrame containing matched data with columns:
        - feature_id_var: identifier column (e.g. feature_id)
        - index_col: index column (e.g. s_id)
        - other columns: data columns to be aggregated
    feature_id_var : str, default="feature_id"
        Name of the identifier column
    index_col : str, default="s_id"
        Name of the column to use as index
    numeric_agg : str, default="weighted_mean"
        Method to aggregate numeric columns:
        - "weighted_mean": weighted by inverse of feature_id frequency (default)
        - "mean": simple arithmetic mean
        - "first": first value after sorting by feature_id_var (requires feature_id_var)
        - "max": maximum value
    keep_id_col : bool, default=True
        Whether to keep and rollup the feature_id_var in the output.
        If False, feature_id_var will be dropped from the output.

    Returns
    -------
    pd.DataFrame
        DataFrame with resolved matches:
        - Many-to-1: numeric columns are aggregated using specified method
        - 1-to-many: adds a count column showing number of matches
        - Index is set to index_col and named accordingly

    Raises
    ------
    KeyError
        If feature_id_var is not present in the DataFrame
    TypeError
        If DataFrame contains unsupported data types (boolean or datetime)
    """
    # Make a copy to avoid modifying input
    df = matched_data.copy()

    # Check for unsupported data types
    unsupported_dtypes = df.select_dtypes(include=["bool", "datetime64"]).columns
    if not unsupported_dtypes.empty:
        raise TypeError(
            f"Unsupported data types found in columns: {list(unsupported_dtypes)}. "
            "Boolean and datetime columns are not supported."
        )

    # Always require feature_id_var
    if feature_id_var not in df.columns:
        raise KeyError(feature_id_var)

    # Deduplicate by feature_id within each s_id using groupby and first BEFORE any further processing
    df = df.groupby([index_col, feature_id_var], sort=False).first().reset_index()

    # Use a unique temporary column name for weights
    if RESOLVE_MATCHES_TMP_WEIGHT_COL in df.columns:
        raise ValueError(
            f"Temporary weight column name '{RESOLVE_MATCHES_TMP_WEIGHT_COL}' already exists in the input data. Please rename or remove this column and try again."
        )

    # Calculate weights if needed (after deduplication!)
    if numeric_agg == RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN:
        feature_counts = df[feature_id_var].value_counts()
        df[RESOLVE_MATCHES_TMP_WEIGHT_COL] = (
            1 / feature_counts[df[feature_id_var]].values
        )

    # Set index for grouping
    df = df.set_index(index_col)

    # Use utility to split columns
    always_non_numeric = [feature_id_var] if keep_id_col else []
    numeric_cols, non_numeric_cols = _split_numeric_non_numeric_columns(
        df, always_non_numeric=always_non_numeric
    )

    # Get aggregator function
    numeric_aggregator = _get_numeric_aggregator(
        method=numeric_agg, feature_id_var=feature_id_var
    )
    resolved = _aggregate_grouped_columns(
        df,
        numeric_cols,
        non_numeric_cols,
        numeric_aggregator,
        feature_id_var=feature_id_var,
        numeric_agg=numeric_agg,
    )
    # Add count of matches per feature_id
    match_counts = matched_data.groupby(index_col)[feature_id_var].nunique()
    resolved[f"{feature_id_var}_match_count"] = match_counts

    # Drop feature_id_var if not keeping it
    if not keep_id_col and feature_id_var in resolved.columns:
        resolved = resolved.drop(columns=[feature_id_var])

    # Ensure index is named consistently
    resolved.index.name = index_col

    return resolved


def _get_wide_results_valid_ontologies(
    results_df: pd.DataFrame, ontologies: Optional[Union[str, list]] = None
) -> list:
    """
    Get the valid ontologies for a wide results dataframe.

    If ontologies is a string, it will be converted to a list.
    If ontologies is None, the column names of the results dataframe which match ONTOLOGIES_LIST will be used.

    Parameters
    ----------
    results_df : pd.DataFrame
        The results dataframe to get the valid ontologies for.
    ontologies : optional str, list
        The ontology to use for the species identifiers. If not provided, the column names of the results dataframes which match ONTOLOGIES_LIST will be used.

    Returns
    -------
    list
        The valid ontologies for the results dataframe.
    """

    if isinstance(ontologies, str):
        ontologies = [ontologies]  # now, it will be None or list

    if ontologies is None:
        ontologies = [col for col in results_df.columns if col in ONTOLOGIES_LIST]
        if len(ontologies) == 0:
            raise ValueError(
                "No valid ontologies found in results dataframe. Columns are: "
                + str(results_df.columns)
            )

    if isinstance(ontologies, list):
        invalid_ontologies = set(ontologies) - set(ONTOLOGIES_LIST)
        if len(invalid_ontologies) > 0:
            raise ValueError(
                "Invalid ontologies found in ontologies list: "
                + str(invalid_ontologies)
            )

    return ontologies


def _get_numeric_aggregator(
    method: str = RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
) -> callable:
    """
    Get aggregation function for numeric columns with various methods.

    Parameters
    ----------
    method : str, default="weighted_mean"
        Aggregation method to use:
        - "weighted_mean": weighted by inverse of feature_id frequency (default)
        - "mean": simple arithmetic mean
        - "first": first value after sorting by feature_id_var (requires feature_id_var)
        - "max": maximum value
    feature_id_var : str, default="feature_id"
        Name of the column specifying a measured feature - used for sorting and weighting

    Returns
    -------
    callable
        Aggregation function to use with groupby

    Raises
    ------
    ValueError
        If method is not recognized
    """

    def weighted_mean(df: pd.DataFrame) -> float:
        # Get values and weights for this group
        values = df["value"]
        weights = df["weight"]
        # Weights are already normalized globally, just use them directly
        return (values * weights).sum() / weights.sum()

    def first_by_id(df: pd.DataFrame) -> float:
        # Sort by feature_id and take first value
        return df.sort_values(feature_id_var).iloc[0]["value"]

    def simple_mean(series: pd.Series) -> float:
        return series.mean()

    def simple_max(series: pd.Series) -> float:
        return series.max()

    aggregators = {
        RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN: weighted_mean,
        RESOLVE_MATCHES_AGGREGATORS.MEAN: simple_mean,
        RESOLVE_MATCHES_AGGREGATORS.FIRST: first_by_id,
        RESOLVE_MATCHES_AGGREGATORS.MAX: simple_max,
    }

    if method not in aggregators:
        raise ValueError(
            f"Unknown aggregation method: {method}. Must be one of {list(aggregators.keys())}"
        )

    return aggregators[method]


def _split_numeric_non_numeric_columns(df: pd.DataFrame, always_non_numeric=None):
    """
    Utility to split DataFrame columns into numeric and non-numeric, always treating specified columns as non-numeric.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to split.
    always_non_numeric : list or set, optional
        Columns to always treat as non-numeric (e.g., ['feature_id']).

    Returns
    -------
    numeric_cols : pd.Index
        Columns considered numeric (int64, float64, and not in always_non_numeric).
    non_numeric_cols : pd.Index
        Columns considered non-numeric (object, string, etc., plus always_non_numeric).
    """
    if always_non_numeric is None:
        always_non_numeric = []
    always_non_numeric = set(always_non_numeric)
    numeric_cols = df.select_dtypes(include=["int64", "float64"]).columns.difference(
        always_non_numeric
    )
    non_numeric_cols = df.columns.difference(numeric_cols)
    return numeric_cols, non_numeric_cols


def _aggregate_grouped_columns(
    df: pd.DataFrame,
    numeric_cols,
    non_numeric_cols,
    numeric_aggregator,
    feature_id_var: str = FEATURE_ID_VAR_DEFAULT,
    numeric_agg: str = RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN,
) -> pd.DataFrame:
    """
    Aggregate numeric and non-numeric columns for grouped DataFrame.
    Assumes deduplication by feature_id within each s_id has already been performed.
    Returns the combined DataFrame.
    """
    results = []

    # Handle non-numeric columns
    if len(non_numeric_cols) > 0:
        non_numeric_agg = (
            df[non_numeric_cols]
            .groupby(level=0)
            .agg(lambda x: ",".join(sorted(set(x.astype(str)))))
        )
        results.append(non_numeric_agg)
    # Handle numeric columns
    if len(numeric_cols) > 0:
        numeric_results = {}
        for col in numeric_cols:
            if numeric_agg in [
                RESOLVE_MATCHES_AGGREGATORS.FIRST,
                RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN,
            ]:
                agg_df = pd.DataFrame(
                    {"value": df[col], feature_id_var: df[feature_id_var]}
                )
                if numeric_agg == RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN:
                    agg_df[RESOLVE_MATCHES_TMP_WEIGHT_COL] = df[
                        RESOLVE_MATCHES_TMP_WEIGHT_COL
                    ]
                numeric_results[col] = agg_df.groupby(level=0).apply(
                    lambda x: (
                        numeric_aggregator(x)
                        if numeric_agg != RESOLVE_MATCHES_AGGREGATORS.WEIGHTED_MEAN
                        else numeric_aggregator(
                            x.rename(columns={RESOLVE_MATCHES_TMP_WEIGHT_COL: "weight"})
                        )
                    )
                )
            else:
                numeric_results[col] = df[col].groupby(level=0).agg(numeric_aggregator)
        numeric_agg_df = pd.DataFrame(numeric_results)
        results.append(numeric_agg_df)
    # Combine results
    if results:
        resolved = pd.concat(results, axis=1)
    else:
        resolved = pd.DataFrame(index=df.index)
    return resolved



from __future__ import annotations

import logging
import os
import re
import shutil
from pydantic import BaseModel
from typing import Optional

from napistu import utils
from napistu.gcs.constants import GCS_ASSETS

from napistu.gcs.constants import INIT_DATA_DIR_MSG
from napistu.gcs.utils import _initialize_data_dir

logger = logging.getLogger(__name__)


def load_public_napistu_asset(
    asset: str,
    data_dir: str,
    subasset: str | None = None,
    init_msg: str = INIT_DATA_DIR_MSG,
    overwrite: bool = False,
) -> str:
    """
    Load Public Napistu Asset

    Download the `asset` asset to `data_dir` if it doesn't
    already exist and return a path

    asset: the file to download (which will be unpacked if its a .tar.gz)
    subasset: the name of a subasset to load from within the asset bundle
    data_dir: the local directory where assets should be stored
    init_msg: message to display if data_dir does not exist
    overwrite: if True, always download the asset and re-extract it, even if it already exists

    returns:
        asset_path: the path to a local file
    """

    # validate data_directory
    _initialize_data_dir(data_dir, init_msg)
    _validate_gcs_asset(asset)
    _validate_gcs_subasset(asset, subasset)

    # get the path for the asset (which may have been downloaded in a tar-ball)
    asset_path = os.path.join(data_dir, _get_gcs_asset_path(asset, subasset))
    if os.path.isfile(asset_path) and not overwrite:
        return asset_path

    download_path = os.path.join(
        data_dir, os.path.basename(GCS_ASSETS.ASSETS[asset]["file"])
    )
    if overwrite:
        _remove_asset_files_if_needed(asset, data_dir)
    if not os.path.isfile(download_path):
        download_public_napistu_asset(asset, download_path)

    # gunzip if needed
    extn = utils.get_extn_from_url(download_path)
    if (
        re.search(".tar\\.gz$", extn)
        or re.search("\\.tgz$", extn)
        or re.search("\\.zip$", extn)
        or re.search("\\.gz$", extn)
    ):
        utils.extract(download_path)

    # check that the asset_path exists
    if not os.path.isfile(asset_path):
        raise FileNotFoundError(
            f"Something went wrong and {asset_path} was not created."
        )

    return asset_path


def download_public_napistu_asset(asset: str, out_path: str) -> None:
    """
    Download Public Napistu Asset

    Args:
        asset (str): The name of a Napistu public asset stored in Google Cloud Storage (GCS)
        out_path (list): Local location where the file should be saved.

    Returns:
        None
    """

    _validate_gcs_asset(asset)
    selected_file = GCS_ASSETS.ASSETS[asset]["public_url"]

    logger.info(f"Downloading {os.path.basename(selected_file)} to {out_path}")
    logger.info(f"Download URI: {selected_file}")

    utils.download_wget(selected_file, out_path)

    if not os.path.isfile(out_path):
        raise FileNotFoundError(f"Download failed: {out_path} was not created.")

    return None


def _validate_gcs_asset(asset: str) -> None:
    """Validate a GCS asset by name."""

    assets = _NapistuAssetsValidator(assets=GCS_ASSETS.ASSETS).assets
    valid_gcs_assets = assets.keys()
    if asset not in valid_gcs_assets:
        raise ValueError(
            f"asset was {asset} and must be one of the keys in GCS_ASSETS.ASSETS: {', '.join(valid_gcs_assets)}"
        )

    return None


def _validate_gcs_subasset(asset: str, subasset: str) -> None:
    """Validate a subasset as belonging to a given asset."""

    if GCS_ASSETS.ASSETS[asset]["subassets"] is None:
        if subasset is not None:
            logger.warning(
                f"subasset was not None but asset {asset} does not have subassets. Ignoring subasset."
            )

        return None

    valid_subassets = GCS_ASSETS.ASSETS[asset]["subassets"]

    if subasset is None:
        raise ValueError(
            f"subasset was None and must be one of {', '.join(valid_subassets)}"
        )

    if subasset not in valid_subassets:
        raise ValueError(
            f"subasset, {subasset}, was not found in asset {asset}. Valid subassets are {', '.join(valid_subassets)}"
        )

    return None


def _get_gcs_asset_path(asset: str, subasset: Optional[str] = None) -> str:
    """
    Get the GCS path for a given asset and subasset.

    Parameters
    ----------
    asset : str
        The name of the asset.
    subasset : Optional[str]
        The name of the subasset.

    Returns
    -------
    str
        The GCS path for the asset or subasset.
    """
    asset_dict = GCS_ASSETS.ASSETS[asset]
    if asset_dict["subassets"] is None:
        out_file = asset_dict["file"]
    else:
        extract_dir = asset_dict["file"].split(".")[0]
        out_file = os.path.join(extract_dir, asset_dict["subassets"][subasset])
    return out_file


class _NapistuAssetValidator(BaseModel):
    file: str
    subassets: dict[str, str] | None
    public_url: str


class _NapistuAssetsValidator(BaseModel):
    assets: dict[str, _NapistuAssetValidator]


def _remove_asset_files_if_needed(asset: str, data_dir: str):
    """
    Remove asset archive and any extracted directory from data_dir.

    Args:
        asset (str): The asset key (e.g., 'test_pathway').
        data_dir (str): The directory where assets are stored.
    """
    logger = logging.getLogger(__name__)
    removed = []

    # Remove the archive file (any extension)
    archive_filename = os.path.basename(GCS_ASSETS.ASSETS[asset]["file"])
    archive_path = os.path.join(data_dir, archive_filename)
    if os.path.exists(archive_path):
        os.remove(archive_path)
        logger.info(f"Removed asset archive: {archive_path}")
        removed.append(archive_path)

    # Remove extracted directory (if any)
    asset_dict = GCS_ASSETS.ASSETS[asset]
    if asset_dict.get("subassets") is not None or any(
        archive_filename.endswith(ext) for ext in [".tar.gz", ".tgz", ".zip", ".gz"]
    ):
        extract_dir = os.path.join(data_dir, archive_filename.split(".")[0])
        if os.path.isdir(extract_dir):
            shutil.rmtree(extract_dir)
            logger.info(f"Removed extracted asset directory: {extract_dir}")
            removed.append(extract_dir)

    if not removed:
        logger.debug("No asset files found to remove.")

    return removed



from __future__ import annotations

import os

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from napistu import sbml_dfs_core
from napistu.ingestion import sbml
from napistu.network import net_create
from napistu.network import net_create_utils
from napistu.network import ng_utils
from napistu.constants import SBML_DFS
from napistu.network.constants import (
    DROP_REACTIONS_WHEN,
    DEFAULT_WT_TRANS,
    WEIGHTING_SPEC,
    GRAPH_WIRING_APPROACHES,
    NAPISTU_GRAPH_EDGES,
    VALID_GRAPH_WIRING_APPROACHES,
)

test_path = os.path.abspath(os.path.join(__file__, os.pardir))
test_data = os.path.join(test_path, "test_data")

sbml_path = os.path.join(test_data, "R-HSA-1237044.sbml")
sbml_model = sbml.SBML(sbml_path)
sbml_dfs = sbml_dfs_core.SBML_dfs(sbml_model)


def test_create_napistu_graph():
    _ = net_create.create_napistu_graph(
        sbml_dfs, wiring_approach=GRAPH_WIRING_APPROACHES.BIPARTITE
    )
    _ = net_create.create_napistu_graph(
        sbml_dfs, wiring_approach=GRAPH_WIRING_APPROACHES.REGULATORY
    )
    _ = net_create.create_napistu_graph(
        sbml_dfs, wiring_approach=GRAPH_WIRING_APPROACHES.SURROGATE
    )


def test_bipartite_regression():
    bipartite_og = net_create.create_napistu_graph(
        sbml_dfs, wiring_approach="bipartite_og"
    )

    bipartite = net_create.create_napistu_graph(
        sbml_dfs, wiring_approach=GRAPH_WIRING_APPROACHES.BIPARTITE
    )

    bipartite_og_edges = bipartite_og.get_edge_dataframe()
    bipartite_edges = bipartite.get_edge_dataframe()

    try:
        pdt.assert_frame_equal(
            bipartite_og_edges, bipartite_edges, check_like=True, check_dtype=False
        )
    except AssertionError as e:
        # Print detailed differences
        print("DataFrames are not equal!")
        print(
            "Shape original:",
            bipartite_og_edges.shape,
            "Shape new:",
            bipartite_edges.shape,
        )
        print(
            "Columns original:",
            bipartite_og_edges.columns.tolist(),
            "Columns new:",
            bipartite_edges.columns.tolist(),
        )
        # Show head of both for quick inspection
        print("Original head:\n", bipartite_og_edges.head())
        print("New head:\n", bipartite_edges.head())
        # Optionally, show where values differ
        if bipartite_og_edges.shape == bipartite_edges.shape:
            diff = bipartite_og_edges != bipartite_edges
            print("Differences (first 5 rows):\n", diff.head())
        raise e  # Re-raise to fail the test


def test_create_napistu_graph_edge_reversed():
    """Test that edge_reversed=True properly reverses edges in the graph for all graph types."""
    # Test each graph type
    for wiring_approach in VALID_GRAPH_WIRING_APPROACHES:
        # Create graphs with and without edge reversal
        normal_graph = net_create.create_napistu_graph(
            sbml_dfs,
            wiring_approach=wiring_approach,
            directed=True,
            edge_reversed=False,
        )
        reversed_graph = net_create.create_napistu_graph(
            sbml_dfs, wiring_approach=wiring_approach, directed=True, edge_reversed=True
        )

        # Get edge dataframes for comparison
        normal_edges = normal_graph.get_edge_dataframe()
        reversed_edges = reversed_graph.get_edge_dataframe()

        # Verify we have edges to test
        assert len(normal_edges) > 0, f"No edges found in {wiring_approach} graph"
        assert len(normal_edges) == len(
            reversed_edges
        ), f"Edge count mismatch in {wiring_approach} graph"

        # Test edge reversal
        # Check a few edges to verify from/to are swapped
        for i in range(min(5, len(normal_edges))):
            # Check from/to are swapped
            assert (
                normal_edges.iloc[i][NAPISTU_GRAPH_EDGES.FROM]
                == reversed_edges.iloc[i][NAPISTU_GRAPH_EDGES.TO]
            ), f"From/to not properly swapped in {wiring_approach} graph"
            assert (
                normal_edges.iloc[i][NAPISTU_GRAPH_EDGES.TO]
                == reversed_edges.iloc[i][NAPISTU_GRAPH_EDGES.FROM]
            ), f"From/to not properly swapped in {wiring_approach} graph"

            # Check stoichiometry is negated
            assert (
                normal_edges.iloc[i][SBML_DFS.STOICHIOMETRY]
                == -reversed_edges.iloc[i][SBML_DFS.STOICHIOMETRY]
            ), f"Stoichiometry not properly negated in {wiring_approach} graph"

            # Check direction attributes are properly swapped
            if normal_edges.iloc[i]["direction"] == "forward":
                assert (
                    reversed_edges.iloc[i]["direction"] == "reverse"
                ), f"Direction not properly reversed (forward->reverse) in {wiring_approach} graph"
            elif normal_edges.iloc[i]["direction"] == "reverse":
                assert (
                    reversed_edges.iloc[i]["direction"] == "forward"
                ), f"Direction not properly reversed (reverse->forward) in {wiring_approach} graph"

            # Check parents/children are swapped
            assert (
                normal_edges.iloc[i]["sc_parents"]
                == reversed_edges.iloc[i]["sc_children"]
            ), f"Parents/children not properly swapped in {wiring_approach} graph"
            assert (
                normal_edges.iloc[i]["sc_children"]
                == reversed_edges.iloc[i]["sc_parents"]
            ), f"Parents/children not properly swapped in {wiring_approach} graph"


def test_create_napistu_graph_none_attrs():
    # Should not raise when reaction_graph_attrs is None
    _ = net_create.create_napistu_graph(
        sbml_dfs, reaction_graph_attrs=None, wiring_approach="bipartite"
    )


def test_process_napistu_graph_none_attrs():
    # Should not raise when reaction_graph_attrs is None
    _ = net_create.process_napistu_graph(sbml_dfs, reaction_graph_attrs=None)


@pytest.mark.skip_on_windows
def test_igraph_loading():
    # test read/write of an igraph network
    directeds = [True, False]
    wiring_approaches = ["bipartite", "regulatory"]

    ng_utils.export_networks(
        sbml_dfs,
        model_prefix="tmp",
        outdir="/tmp",
        directeds=directeds,
        wiring_approaches=wiring_approaches,
    )

    for wiring_approach in wiring_approaches:
        for directed in directeds:
            import_pkl_path = ng_utils._create_network_save_string(
                model_prefix="tmp",
                outdir="/tmp",
                directed=directed,
                wiring_approach=wiring_approach,
            )
            network_graph = ng_utils.read_network_pkl(
                model_prefix="tmp",
                network_dir="/tmp",
                directed=directed,
                wiring_approach=wiring_approach,
            )

            assert network_graph.is_directed() == directed
            # cleanup
            os.unlink(import_pkl_path)


def test_reverse_network_edges(reaction_species_examples):

    graph_hierarchy_df = net_create_utils.create_graph_hierarchy_df("regulatory")

    rxn_edges = net_create_utils.format_tiered_reaction_species(
        rxn_species=reaction_species_examples["all_entities"],
        r_id="foo",
        graph_hierarchy_df=graph_hierarchy_df,
        drop_reactions_when=DROP_REACTIONS_WHEN.SAME_TIER,
    )
    augmented_network_edges = rxn_edges.assign(r_isreversible=True)
    augmented_network_edges["sc_parents"] = range(0, augmented_network_edges.shape[0])
    augmented_network_edges["sc_children"] = range(
        augmented_network_edges.shape[0], 0, -1
    )

    assert net_create._reverse_network_edges(augmented_network_edges).shape[0] == 2


def test_entity_validation():
    # Test basic validation
    entity_attrs = {"table": "reactions", "variable": "foo"}
    assert net_create._EntityAttrValidator(**entity_attrs).model_dump() == {
        **entity_attrs,
        **{"trans": DEFAULT_WT_TRANS},
    }

    # Test validation with custom transformations
    custom_transformations = {
        "nlog10": lambda x: -np.log10(x),
        "square": lambda x: x**2,
    }

    # Test valid custom transformation
    entity_attrs_custom = {
        "attr1": {
            WEIGHTING_SPEC.TABLE: "reactions",
            WEIGHTING_SPEC.VARIABLE: "foo",
            WEIGHTING_SPEC.TRANSFORMATION: "nlog10",
        },
        "attr2": {
            WEIGHTING_SPEC.TABLE: "species",
            WEIGHTING_SPEC.VARIABLE: "bar",
            WEIGHTING_SPEC.TRANSFORMATION: "square",
        },
    }
    # Should not raise any errors
    net_create._validate_entity_attrs(
        entity_attrs_custom, custom_transformations=custom_transformations
    )

    # Test invalid transformation
    entity_attrs_invalid = {
        "attr1": {
            WEIGHTING_SPEC.TABLE: "reactions",
            WEIGHTING_SPEC.VARIABLE: "foo",
            WEIGHTING_SPEC.TRANSFORMATION: "invalid_trans",
        }
    }
    with pytest.raises(ValueError) as excinfo:
        net_create._validate_entity_attrs(
            entity_attrs_invalid, custom_transformations=custom_transformations
        )
    assert "transformation 'invalid_trans' was not defined" in str(excinfo.value)

    # Test with validate_transformations=False
    # Should not raise any errors even with invalid transformation
    net_create._validate_entity_attrs(
        entity_attrs_invalid, validate_transformations=False
    )

    # Test with non-dict input
    with pytest.raises(AssertionError) as excinfo:
        net_create._validate_entity_attrs(["not", "a", "dict"])
    assert "entity_attrs must be a dictionary" in str(excinfo.value)


def test_pluck_entity_data_species_identity(sbml_dfs):
    # Take first 10 species IDs
    species_ids = sbml_dfs.species.index[:10]
    # Create mock data with explicit dtype to ensure cross-platform consistency
    # Fix for issue-42: Use explicit dtypes to avoid platform-specific dtype differences
    # between Windows (int32) and macOS/Linux (int64)
    mock_df = pd.DataFrame(
        {
            "string_col": [f"str_{i}" for i in range(10)],
            "mixed_col": np.arange(-5, 5, dtype=np.int64),  # Explicitly use int64
            "ones_col": np.ones(10, dtype=np.float64),  # Explicitly use float64
            "squared_col": np.arange(10, dtype=np.int64),  # Explicitly use int64
        },
        index=species_ids,
    )
    # Assign to species_data
    sbml_dfs.species_data["mock_table"] = mock_df

    # Custom transformation: square
    def square(x):
        return x**2

    custom_transformations = {"square": square}
    # Create graph_attrs for species
    graph_attrs = {
        "species": {
            "string_col": {
                WEIGHTING_SPEC.TABLE: "mock_table",
                WEIGHTING_SPEC.VARIABLE: "string_col",
                WEIGHTING_SPEC.TRANSFORMATION: "identity",
            },
            "mixed_col": {
                WEIGHTING_SPEC.TABLE: "mock_table",
                WEIGHTING_SPEC.VARIABLE: "mixed_col",
                WEIGHTING_SPEC.TRANSFORMATION: "identity",
            },
            "ones_col": {
                WEIGHTING_SPEC.TABLE: "mock_table",
                WEIGHTING_SPEC.VARIABLE: "ones_col",
                WEIGHTING_SPEC.TRANSFORMATION: "identity",
            },
            "squared_col": {
                WEIGHTING_SPEC.TABLE: "mock_table",
                WEIGHTING_SPEC.VARIABLE: "squared_col",
                WEIGHTING_SPEC.TRANSFORMATION: "square",
            },
        }
    }
    # Call pluck_entity_data with custom transformation
    result = net_create.pluck_entity_data(
        sbml_dfs, graph_attrs, "species", custom_transformations=custom_transformations
    )
    # Check output
    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == {"string_col", "mixed_col", "ones_col", "squared_col"}
    assert list(result.index) == list(species_ids)
    # Check values
    pd.testing.assert_series_equal(result["string_col"], mock_df["string_col"])
    pd.testing.assert_series_equal(result["mixed_col"], mock_df["mixed_col"])
    pd.testing.assert_series_equal(result["ones_col"], mock_df["ones_col"])
    pd.testing.assert_series_equal(
        result["squared_col"], mock_df["squared_col"].apply(square)
    )


def test_pluck_entity_data_missing_species_key(sbml_dfs):
    # graph_attrs does not contain 'species' key
    graph_attrs = {}
    result = net_create.pluck_entity_data(sbml_dfs, graph_attrs, SBML_DFS.SPECIES)
    assert result is None


def test_pluck_entity_data_empty_species_dict(sbml_dfs):
    # graph_attrs contains 'species' key but value is empty dict
    graph_attrs = {SBML_DFS.SPECIES: {}}
    result = net_create.pluck_entity_data(sbml_dfs, graph_attrs, SBML_DFS.SPECIES)
    assert result is None



from __future__ import annotations

import copy
import logging
from typing import Any, Optional

import igraph as ig
import pandas as pd

from napistu.network.constants import (
    NAPISTU_GRAPH_EDGES,
    EDGE_REVERSAL_ATTRIBUTE_MAPPING,
    EDGE_DIRECTION_MAPPING,
)

logger = logging.getLogger(__name__)


class NapistuGraph(ig.Graph):
    """
    A subclass of igraph.Graph with additional functionality for molecular network analysis.

    This class extends igraph.Graph with domain-specific methods and metadata tracking
    for biological pathway and molecular interaction networks. All standard igraph
    methods are available, plus additional functionality for edge reversal and
    metadata management.

    Parameters
    ----------
    *args : tuple
        Positional arguments passed to igraph.Graph constructor
    **kwargs : dict
        Keyword arguments passed to igraph.Graph constructor

    Attributes
    ----------
    is_reversed : bool
        Whether the graph edges have been reversed from their original direction
    wiring_approach : str or None
        Type of graph (e.g., 'bipartite', 'regulatory', 'surrogate')
    weighting_strategy : str or None
        Strategy used for edge weighting (e.g., 'topology', 'mixed', 'calibrated')

    Methods
    -------
    from_igraph(graph, **metadata)
        Create a NapistuGraph from an existing igraph.Graph
    reverse_edges()
        Reverse all edges in the graph in-place
    set_metadata(**kwargs)
        Set metadata for the graph in-place
    get_metadata(key=None)
        Get metadata from the graph
    copy()
        Create a deep copy of the NapistuGraph

    Examples
    --------
    Create a NapistuGraph from scratch:

    >>> ng = NapistuGraph(directed=True)
    >>> ng.add_vertices(3)
    >>> ng.add_edges([(0, 1), (1, 2)])

    Convert from existing igraph:

    >>> import igraph as ig
    >>> g = ig.Graph.Erdos_Renyi(10, 0.3)
    >>> ng = NapistuGraph.from_igraph(g, graph_type='random')

    Reverse edges and check state:

    >>> ng.reverse_edges()
    >>> print(ng.is_reversed)
    True

    Set and retrieve metadata:

    >>> ng.set_metadata(experiment_id='exp_001', date='2024-01-01')
    >>> print(ng.get_metadata('experiment_id'))
    'exp_001'

    Notes
    -----
    NapistuGraph inherits from igraph.Graph, so all standard igraph methods
    (degree, shortest_paths, betweenness, etc.) are available. The additional
    functionality is designed specifically for molecular network analysis.

    Edge reversal swaps 'from'/'to' attributes, negates stoichiometry values,
    and updates direction metadata according to predefined mapping rules.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize a NapistuGraph.

        Accepts all the same arguments as igraph.Graph constructor.
        """
        super().__init__(*args, **kwargs)

        # Initialize metadata
        self._metadata = {
            "is_reversed": False,
            "wiring_approach": None,
            "weighting_strategy": None,
            "creation_params": {},
        }

    @classmethod
    def from_igraph(cls, graph: ig.Graph, **metadata) -> "NapistuGraph":
        """
        Create a NapistuGraph from an existing igraph.Graph.

        Parameters
        ----------
        graph : ig.Graph
            The igraph to convert
        **metadata : dict
            Additional metadata to store with the graph

        Returns
        -------
        NapistuGraph
            A new NapistuGraph instance
        """
        # Create new instance with same structure
        new_graph = cls(
            n=graph.vcount(),
            edges=[(e.source, e.target) for e in graph.es],
            directed=graph.is_directed(),
        )

        # Copy all vertex attributes
        for attr in graph.vs.attributes():
            new_graph.vs[attr] = graph.vs[attr]

        # Copy all edge attributes
        for attr in graph.es.attributes():
            new_graph.es[attr] = graph.es[attr]

        # Copy graph attributes
        for attr in graph.attributes():
            new_graph[attr] = graph[attr]

        # Set metadata
        new_graph._metadata.update(metadata)

        return new_graph

    @property
    def is_reversed(self) -> bool:
        """Check if the graph has been reversed."""
        return self._metadata["is_reversed"]

    @property
    def wiring_approach(self) -> Optional[str]:
        """Get the graph type (bipartite, regulatory, etc.)."""
        return self._metadata["wiring_approach"]

    @property
    def weighting_strategy(self) -> Optional[str]:
        """Get the weighting strategy used."""
        return self._metadata["weighting_strategy"]

    def reverse_edges(self) -> None:
        """
        Reverse all edges in the graph.

        This swaps edge directions and updates all associated attributes
        according to the edge reversal mapping utilities. Modifies the graph in-place.

        Returns
        -------
        None
        """
        # Get current edge dataframe
        edges_df = self.get_edge_dataframe()

        # Apply systematic attribute swapping using utilities
        reversed_edges_df = _apply_edge_reversal_mapping(edges_df)

        # Handle special cases using utilities
        reversed_edges_df = _handle_special_reversal_cases(reversed_edges_df)

        # Update edge attributes
        for attr in reversed_edges_df.columns:
            if attr in self.es.attributes():
                self.es[attr] = reversed_edges_df[attr].values

        # Update metadata
        self._metadata["is_reversed"] = not self._metadata["is_reversed"]

        logger.info(
            f"Reversed graph edges. Current state: reversed={self._metadata['is_reversed']}"
        )

        return None

    def remove_isolated_vertices(self):
        """
        Remove vertices that have no edges (degree 0) from the graph.


        Returns
        -------
        None
            The graph is modified in-place.

        """

        # Find isolated vertices (degree 0)
        isolated_vertices = self.vs.select(_degree=0)

        if len(isolated_vertices) == 0:
            logger.info("No isolated vertices found to remove")
            return

        # Get vertex names/indices for logging (up to 5 examples)
        vertex_names = []
        for v in isolated_vertices[:5]:
            # Use vertex name if available, otherwise use index
            name = (
                v["name"]
                if "name" in v.attributes() and v["name"] is not None
                else str(v.index)
            )
            vertex_names.append(name)

        # Create log message
        examples_str = ", ".join(f"'{name}'" for name in vertex_names)
        if len(isolated_vertices) > 5:
            examples_str += f" (and {len(isolated_vertices) - 5} more)"

        logger.info(
            f"Removed {len(isolated_vertices)} isolated vertices: [{examples_str}]"
        )

        # Remove the isolated vertices
        self.delete_vertices(isolated_vertices)

    def set_metadata(self, **kwargs) -> None:
        """
        Set metadata for the graph.

        Modifies the graph's metadata in-place.

        Parameters
        ----------
        **kwargs : dict
            Metadata key-value pairs to set
        """
        self._metadata.update(kwargs)

        return None

    def get_metadata(self, key: Optional[str] = None) -> Any:
        """
        Get metadata from the graph.

        Parameters
        ----------
        key : str, optional
            Specific metadata key to retrieve. If None, returns all metadata.

        Returns
        -------
        Any
            The requested metadata value, or all metadata if key is None
        """
        if key is None:
            return self._metadata.copy()
        return self._metadata.get(key)

    def copy(self) -> "NapistuGraph":
        """
        Create a deep copy of the NapistuGraph.

        Returns
        -------
        NapistuGraph
            A deep copy of this graph including metadata
        """
        # Use igraph's copy method to get the graph structure and attributes
        new_graph = super().copy()

        # Convert to NapistuGraph and copy metadata
        napistu_copy = NapistuGraph.from_igraph(new_graph)
        napistu_copy._metadata = copy.deepcopy(self._metadata)

        return napistu_copy

    def __str__(self) -> str:
        """String representation including metadata."""
        base_str = super().__str__()
        metadata_str = (
            f"Reversed: {self.is_reversed}, "
            f"Type: {self.wiring_approach}, "
            f"Weighting: {self.weighting_strategy}"
        )
        return f"{base_str}\nNapistuGraph metadata: {metadata_str}"

    def __repr__(self) -> str:
        """Detailed representation."""
        return self.__str__()


def _apply_edge_reversal_mapping(edges_df: pd.DataFrame) -> pd.DataFrame:
    """
    Apply systematic attribute mapping for edge reversal.

    This function swaps paired attributes according to EDGE_REVERSAL_ATTRIBUTE_MAPPING.
    For example, 'from' becomes 'to', 'weights' becomes 'upstream_weights', etc.

    Parameters
    ----------
    edges_df : pd.DataFrame
        Current edge attributes

    Returns
    -------
    pd.DataFrame
        Edge dataframe with swapped attributes

    Warnings
    --------
    Logs warnings when expected attribute pairs are missing
    """
    # Find which attributes have pairs in the mapping
    available_attrs = set(edges_df.columns)

    # Find pairs where both attributes exist
    valid_mapping = {}
    missing_pairs = []

    for source_attr, target_attr in EDGE_REVERSAL_ATTRIBUTE_MAPPING.items():
        if source_attr in available_attrs:
            if target_attr in available_attrs:
                valid_mapping[source_attr] = target_attr
            else:
                missing_pairs.append(f"{source_attr} -> {target_attr}")

    # Warn about attributes that can't be swapped
    if missing_pairs:
        logger.warning(
            f"The following edge attributes cannot be swapped during reversal "
            f"because their paired attribute is missing: {', '.join(missing_pairs)}"
        )

    return edges_df.rename(columns=valid_mapping)


def _handle_special_reversal_cases(edges_df: pd.DataFrame) -> pd.DataFrame:
    """
    Handle special cases that need more than simple attribute swapping.

    This includes:
    - Flipping stoichiometry signs (* -1)
    - Mapping direction enums (forward <-> reverse)

    Parameters
    ----------
    edges_df : pd.DataFrame
        Edge dataframe after basic attribute swapping

    Returns
    -------
    pd.DataFrame
        Edge dataframe with special cases handled

    Warnings
    --------
    Logs warnings when expected attributes are missing
    """
    result_df = edges_df.copy()

    # Handle stoichiometry sign flip
    if NAPISTU_GRAPH_EDGES.STOICHIOMETRY in result_df.columns:
        result_df[NAPISTU_GRAPH_EDGES.STOICHIOMETRY] *= -1
    else:
        logger.warning(
            f"Missing expected '{NAPISTU_GRAPH_EDGES.STOICHIOMETRY}' attribute during edge reversal. "
            "Stoichiometry signs will not be flipped."
        )

    # Handle direction enum mapping
    if NAPISTU_GRAPH_EDGES.DIRECTION in result_df.columns:
        result_df[NAPISTU_GRAPH_EDGES.DIRECTION] = result_df[
            NAPISTU_GRAPH_EDGES.DIRECTION
        ].map(EDGE_DIRECTION_MAPPING)
    else:
        logger.warning(
            f"Missing expected '{NAPISTU_GRAPH_EDGES.DIRECTION}' attribute during edge reversal. "
            "Direction metadata will not be updated."
        )

    return result_df



from __future__ import annotations

import pandas as pd
from napistu import indices
from napistu.constants import SOURCE_SPEC


class Source:
    """
    An Entity's Source

    Attributes
    ----------
    source : pd.DataFrame
        A dataframe containing the model source and other optional variables

    Methods
    -------

    """

    def __init__(
        self,
        source_df: pd.DataFrame | None = None,
        init: bool = False,
        pw_index: indices.PWIndex | None = None,
    ) -> None:
        """
        Tracks the model(s) an entity (i.e., a compartment, species, reaction) came from.

        By convention sources exist only for the models that an entity came from rather
        than the current model they are part of. For example, when combining Reactome models
        into a consensus, a molecule which existed in multiple models would have a source entry
        for each, but it would not have a source entry for the consensus model itself.

        Parameters
        ----------
        source_df : pd.DataFrame
            A dataframe containing the model source and other optional variables
        init : bool
            Creates an empty source object. This is typically used when creating an SBML_dfs
            object from a single source.
        pw_index : indices.PWIndex

        Returns
        -------
        None.

        """

        if init is True:
            # initialize with an empty Source
            self.source = None
        else:
            if isinstance(source_df, pd.DataFrame):
                # if pw_index is provided then it will be joined to source_df to add additional metadata
                if pw_index is not None:
                    if not isinstance(pw_index, indices.PWIndex):
                        raise ValueError(
                            f"pw_index must be a indices.PWIndex or None and was {type(pw_index).__name__}"
                        )
                    else:
                        # check that all models are present in the pathway index
                        missing_pathways = set(
                            source_df[SOURCE_SPEC.MODEL].tolist()
                        ).difference(
                            set(pw_index.index[SOURCE_SPEC.PATHWAY_ID].tolist())
                        )
                        if len(missing_pathways) > 0:
                            raise ValueError(
                                f"{len(missing_pathways)} pathway models are present"
                                f" in source_df but not the pw_index: {', '.join(missing_pathways)}"
                            )

                        source_df = source_df.merge(
                            pw_index.index,
                            left_on=SOURCE_SPEC.MODEL,
                            right_on=SOURCE_SPEC.PATHWAY_ID,
                        )

                self.source = source_df
            else:
                raise TypeError(
                    'source_df must be a pd.DataFrame if "init" is False, but was type '
                    f"{type(source_df).__name__}"
                )

            if SOURCE_SPEC.MODEL not in source_df.columns.values.tolist():
                raise ValueError(
                    f"{SOURCE_SPEC.MODEL} variable was not found, but is required in a Source object"
                )
            if SOURCE_SPEC.PATHWAY_ID not in source_df.columns.values.tolist():
                raise ValueError(
                    f"{SOURCE_SPEC.PATHWAY_ID} variable was not found, but is required in a Source object"
                )


def create_source_table(
    lookup_table: pd.Series, table_schema: dict, pw_index: indices.PWIndex | None
) -> pd.DataFrame:
    """
    Create Source Table

    Create a table with one row per "new_id" and a Source object created from the union
      of "old_id" Source objects
    """

    if SOURCE_SPEC.SOURCE not in table_schema.keys():
        raise ValueError(
            f"{SOURCE_SPEC.SOURCE} not present in schema, can't create source_table"
        )

    # take lookup_table and create an index on "new_id". Multiple rows may have the
    # same value for new_id so these are grouped together.
    lookup_table_rearranged = lookup_table.reset_index().set_index(["new_id"])

    # run a list comprehension over each value of new_id to create a Source
    # object based on the dataframe specific to new_id
    # pw_index is provided to fill out additional meta-information beyond the
    # pathway_id which defines a single source
    def create_source(group):
        return Source(
            group.reset_index(drop=True),
            pw_index=pw_index,
        )

    id_table = (
        lookup_table_rearranged.groupby("new_id")
        .apply(create_source)
        .rename(table_schema[SOURCE_SPEC.SOURCE])
        .to_frame()
    )

    id_table.index = id_table.index.rename(table_schema["pk"])

    return id_table


def merge_sources(source_list: list | pd.Series) -> Source:
    """
    Merge Sources

    Merge a list of Source objects into a single Source object

    """

    # filter to non-empty sources
    # empty sources have only been initialized; a merge hasn't occured
    existing_sources = [s.source is not None for s in source_list]
    if not any(existing_sources):
        if isinstance(source_list, list):
            return source_list[0]
        else:
            return source_list.iloc[0]

    existing_source_list = [
        x.source for x, y in zip(source_list, existing_sources) if y
    ]

    return Source(pd.concat(existing_source_list))


def unnest_sources(
    source_table: pd.DataFrame, source_var: str, verbose: bool = False
) -> pd.DataFrame:
    """
    Unnest Sources

    Take a pd.DataFrame containing an array of Sources and
    return one-row per source.

    Parameters:
    source_table: pd.DataFrame
        a table containing an array of Sources
    source_var: str
        variable containing Sources

    Returns:
    pd.Dataframe containing the index of source_table but expanded
    to include one row per source

    """

    sources = list()
    source_table_index = source_table.index.to_frame().reset_index(drop=True)

    for i in range(source_table.shape[0]):
        if verbose:
            print(f"Processing {source_table_index.index.values[i]}")

        # check that the entries of sourcevar are Source objects
        source_value = source_table[source_var].iloc[i]

        if not isinstance(source_value, Source):
            raise TypeError(
                f"source_value must be a Source, but got {type(source_value).__name__}"
            )

        if source_value.source is None:
            print("Some sources were only missing - returning None")
            return None

        source_tbl = pd.DataFrame(source_value.source)
        source_tbl.index.name = SOURCE_SPEC.INDEX_NAME
        source_tbl = source_tbl.reset_index()

        # add original index as variables and then set index
        for j in range(source_table_index.shape[1]):
            source_tbl[source_table_index.columns[j]] = source_table_index.iloc[i, j]
        source_tbl = source_tbl.set_index(
            list(source_table_index.columns) + [SOURCE_SPEC.INDEX_NAME]
        )

        sources.append(source_tbl)

    return pd.concat(sources)


def greedy_set_coverge_of_sources(
    source_df: pd.DataFrame, table_schema: dict
) -> pd.DataFrame:
    """
    Greedy Set Coverage of Sources

    Apply the greedy set coverge algorithm to find the minimal set of
    sources which cover all entries

    Parameters:
    source_df: pd.DataFrame
        pd.Dataframe containing the index of source_table but expanded to
        include one row per source. As produced by source.unnest_sources()

    Returns:
    minimial_sources: [str]
        A list of pathway_ids of the minimal source set

    """

    # rollup pathways with identical membership
    deduplicated_sources = _deduplicate_source_df(source_df, table_schema)

    unaccounted_for_members = deduplicated_sources
    retained_pathway_ids = []

    while unaccounted_for_members.shape[0] != 0:
        # find the pathway with the most members
        pathway_members = unaccounted_for_members.groupby(SOURCE_SPEC.PATHWAY_ID).size()
        top_pathway = pathway_members[pathway_members == max(pathway_members)].index[0]
        retained_pathway_ids.append(top_pathway)

        # remove all members associated with the top pathway
        members_captured = (
            unaccounted_for_members[
                unaccounted_for_members[SOURCE_SPEC.PATHWAY_ID] == top_pathway
            ]
            .index.get_level_values(table_schema["pk"])
            .tolist()
        )

        unaccounted_for_members = unaccounted_for_members[
            ~unaccounted_for_members.index.get_level_values(table_schema["pk"]).isin(
                members_captured
            )
        ]

    minimial_sources = deduplicated_sources[
        deduplicated_sources[SOURCE_SPEC.PATHWAY_ID].isin(retained_pathway_ids)
    ].sort_index()

    return minimial_sources


def _deduplicate_source_df(source_df: pd.DataFrame, table_schema: dict) -> pd.DataFrame:
    """Combine entries in a source table when multiple models have the same members."""

    # drop entries which are missing required attributes and throw an error if none are left
    REQUIRED_NON_NA_ATTRIBUTES = [SOURCE_SPEC.PATHWAY_ID]
    indexed_sources = (
        source_df.reset_index()
        .merge(source_df[REQUIRED_NON_NA_ATTRIBUTES].dropna())
        .set_index(SOURCE_SPEC.PATHWAY_ID)
    )

    if indexed_sources.shape[0] == 0:
        raise ValueError(
            f"source_df was provided but zero entries had a defined {' OR '.join(REQUIRED_NON_NA_ATTRIBUTES)}"
        )

    pathways = indexed_sources.index.unique()

    # identify pathways with identical coverage

    pathway_member_string = (
        pd.DataFrame(
            [
                {
                    SOURCE_SPEC.PATHWAY_ID: p,
                    "membership_string": "_".join(
                        set(indexed_sources.loc[[p]][table_schema["pk"]].tolist())
                    ),
                }
                for p in pathways
            ]
        )
        .drop_duplicates()
        .set_index("membership_string")
    )

    membership_categories = pathway_member_string.merge(
        source_df.groupby(SOURCE_SPEC.PATHWAY_ID).first(),
        left_on=SOURCE_SPEC.PATHWAY_ID,
        right_index=True,
    )

    category_index = membership_categories.index.unique()
    if not isinstance(category_index, pd.core.indexes.base.Index):
        raise TypeError(
            f"category_index must be a pandas Index, but got {type(category_index).__name__}"
        )

    merged_sources = pd.concat(
        [
            _collapse_by_membership_string(s, membership_categories, table_schema)  # type: ignore
            for s in category_index.tolist()
        ]
    )
    merged_sources[SOURCE_SPEC.INDEX_NAME] = merged_sources.groupby(
        table_schema["pk"]
    ).cumcount()

    return merged_sources.set_index(
        [table_schema["pk"], SOURCE_SPEC.INDEX_NAME]
    ).sort_index()


def _collapse_by_membership_string(
    membership_string: str, membership_categories: pd.DataFrame, table_schema: dict
) -> pd.DataFrame:
    """Assign each member of a membership-string to a set of pathways."""

    collapsed_source_membership = _collapse_source_df(
        membership_categories.loc[membership_string]
    )

    return pd.DataFrame(
        [
            pd.concat(
                [pd.Series({table_schema["pk"]: ms}), collapsed_source_membership]
            )
            for ms in membership_string.split("_")
        ]
    )


def _collapse_source_df(source_df: pd.DataFrame) -> pd.Series:
    """Collapse a source_df table into a single entry."""

    if isinstance(source_df, pd.DataFrame):
        collapsed_source_series = pd.Series(
            {
                SOURCE_SPEC.PATHWAY_ID: " OR ".join(source_df[SOURCE_SPEC.PATHWAY_ID]),
                SOURCE_SPEC.MODEL: " OR ".join(source_df[SOURCE_SPEC.MODEL]),
                SOURCE_SPEC.SOURCE: " OR ".join(
                    set(source_df[SOURCE_SPEC.SOURCE].tolist())
                ),
                SOURCE_SPEC.SPECIES: " OR ".join(
                    set(source_df[SOURCE_SPEC.SPECIES].tolist())
                ),
                SOURCE_SPEC.NAME: " OR ".join(source_df[SOURCE_SPEC.NAME]),
                SOURCE_SPEC.N_COLLAPSED_PATHWAYS: source_df.shape[0],
            }
        )
    elif isinstance(source_df, pd.Series):
        collapsed_source_series = pd.Series(
            {
                SOURCE_SPEC.PATHWAY_ID: source_df[SOURCE_SPEC.PATHWAY_ID],
                SOURCE_SPEC.MODEL: source_df[SOURCE_SPEC.MODEL],
                SOURCE_SPEC.SOURCE: source_df[SOURCE_SPEC.SOURCE],
                SOURCE_SPEC.SPECIES: source_df[SOURCE_SPEC.SPECIES],
                SOURCE_SPEC.NAME: source_df[SOURCE_SPEC.NAME],
                SOURCE_SPEC.N_COLLAPSED_PATHWAYS: 1,
            }
        )
    else:
        raise TypeError(
            f"source_df must be a pd.DataFrame or pd.Series, but was a {type(source_df).__name__}"
        )

    return collapsed_source_series


def _safe_source_merge(member_Sources: Source | list) -> Source:
    """Combine either a Source or pd.Series of Sources into a single Source object."""

    if isinstance(member_Sources, Source):
        return member_Sources
    elif isinstance(member_Sources, pd.Series):
        return merge_sources(member_Sources.tolist())
    else:
        raise TypeError("Expecting source.Source or pd.Series")



from types import SimpleNamespace

FEATURE_ID_VAR_DEFAULT = "feature_id"

RESOLVE_MATCHES_AGGREGATORS = SimpleNamespace(
    WEIGHTED_MEAN="weighted_mean", MEAN="mean", FIRST="first", MAX="max"
)

RESOLVE_MATCHES_TMP_WEIGHT_COL = "__tmp_weight_for_aggregation__"

BIND_DICT_OF_WIDE_RESULTS_STRATEGIES = SimpleNamespace(
    CONTATENATE="concatenate", MULTIPLE_KEYS="multiple_keys", STAGGER="stagger"
)
BIND_DICT_OF_WIDE_RESULTS_STRATEGIES_LIST = [
    BIND_DICT_OF_WIDE_RESULTS_STRATEGIES.CONTATENATE,
    BIND_DICT_OF_WIDE_RESULTS_STRATEGIES.MULTIPLE_KEYS,
    BIND_DICT_OF_WIDE_RESULTS_STRATEGIES.STAGGER,
]



from __future__ import annotations

import os

import pandas as pd
from napistu import sbml_dfs_core
from napistu.ingestion import sbml
from napistu.modify import curation

test_path = os.path.abspath(os.path.join(__file__, os.pardir))
sbml_path = os.path.join(test_path, "test_data", "R-HSA-1237044.sbml")

if not os.path.isfile(sbml_path):
    raise ValueError(f"{sbml_path} not found")

# setup mock curations
curation_dict = dict()
curation_dict["species"] = pd.DataFrame(
    [
        {
            "species": "hello",
            "uri": "http://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:35828",
            "curator": "Sean",
        },
        {"species": "good day", "uri": None, "curator": "Sean"},
    ]
)

curation_dict["compartmentalized_species"] = pd.DataFrame(
    [
        {
            "compartmentalized_species": "hello [cytosol]",
            "s_name": "hello",
            "c_name": "cytosol",
            "curator": "Sean",
        }
    ]
)

curation_dict["reactions"] = pd.DataFrame(
    [
        {
            "reactions": "there",
            "stoichiometry": "hello [cytosol] -> CO2 [cytosol]",
            "uri": None,
            "evidence": "how is",
            "curator": "Sean",
        },
        {
            "reactions": "where",
            "stoichiometry": "CO2 [cytosol] -> hello [cytosol]",
            "uri": None,
            "evidence": "your family",
            "curator": "Sean",
        },
    ]
)

curation_dict["reaction_species"] = pd.DataFrame(
    [
        {
            "reaction_species": "NADH [cytosol]",
            "r_name": "CYB5Rs reduce MetHb to HbA",
            "stoichiometry": 0,
            "sbo_term_name": "stimulator",
            "evidence": "weeeee",
            "curator": "Sean",
        }
    ]
)

curation_dict["remove"] = pd.DataFrame(
    [
        {"remove": "reaction_1237042", "table": "reactions", "variable": "r_id"},
        {
            "remove": "CYB5Rs reduce MetHb to HbA",
            "table": "reactions",
            "variable": "r_name",
        },
        {"remove": "CO2", "table": "species", "variable": "s_name"},
    ]
)


def test_remove_entities():
    sbml_model = sbml.SBML(sbml_path)
    sbml_dfs = sbml_dfs_core.SBML_dfs(sbml_model)
    sbml_dfs.validate()

    invalid_entities_dict = curation._find_invalid_entities(
        sbml_dfs, curation_dict["remove"]
    )
    invalid_pks = set(invalid_entities_dict.keys())

    assert invalid_pks == {"sc_id", "rsc_id", "r_id", "s_id"}

    n_species = sbml_dfs.species.shape[0]
    n_reactions = sbml_dfs.reactions.shape[0]
    n_compartmentalized_species = sbml_dfs.compartmentalized_species.shape[0]
    n_reaction_species = sbml_dfs.reaction_species.shape[0]
    # should be untouched
    n_compartments = sbml_dfs.compartments.shape[0]

    sbml_dfs = curation._remove_entities(sbml_dfs, invalid_entities_dict)

    assert n_species - sbml_dfs.species.shape[0] == 1
    assert n_reactions - sbml_dfs.reactions.shape[0] == 2
    assert (
        n_compartmentalized_species - sbml_dfs.compartmentalized_species.shape[0] == 2
    )
    assert n_reaction_species - sbml_dfs.reaction_species.shape[0] == 14
    assert n_compartments - sbml_dfs.compartments.shape[0] == 0


def test_add_entities():
    sbml_model = sbml.SBML(sbml_path)
    sbml_dfs = sbml_dfs_core.SBML_dfs(sbml_model)
    sbml_dfs.validate()

    new_entities = curation.format_curations(curation_dict, sbml_dfs)

    assert new_entities["species"].shape == (2, 3)
    assert new_entities["reactions"].shape == (2, 4)
    assert new_entities["compartmentalized_species"].shape == (1, 4)
    assert new_entities["reaction_species"].shape == (5, 4)


################################################
# __main__
################################################

if __name__ == "__main__":
    test_remove_entities()
    test_add_entities()



"""
Utilities for loading and processing tutorials.
"""

import os
from pathlib import Path
from typing import Dict
import httpx
import logging

from napistu.gcs.utils import _initialize_data_dir

from napistu.mcp.constants import TUTORIAL_URLS
from napistu.mcp.constants import TUTORIALS_CACHE_DIR

logger = logging.getLogger(__name__)

# Import optional dependencies with error handling
try:
    import nbformat
except ImportError:
    raise ImportError(
        "Tutorial utilities require additional dependencies. Install with 'pip install napistu[mcp]'"
    )

# Configure logger for this module
logger = logging.getLogger(__name__)


async def get_tutorial_markdown(
    tutorial_id: str,
    tutorial_urls: Dict[str, str] = TUTORIAL_URLS,
    cache_dir: Path = TUTORIALS_CACHE_DIR,
) -> str:
    """
    Download/cache the notebook if needed, load it, and return the markdown.

    Parameters
    ----------
    tutorial_id : str
        The ID of the tutorial (key in tutorial_urls).
    tutorial_urls : dict, optional
        Mapping of tutorial IDs to GitHub raw URLs. Defaults to TUTORIAL_URLS.
    cache_dir : Path, optional
        Directory to cache downloaded notebooks. Defaults to TUTORIALS_CACHE_DIR.

    Returns
    -------
    str
        Markdown content of the notebook as a string.

    Raises
    ------
    Exception
        If the notebook cannot be downloaded, loaded, or parsed.

    Examples
    --------
    >>> markdown = await get_tutorial_markdown('my_tutorial')
    >>> print(markdown)
    """
    try:
        path = await _ensure_notebook_cached(tutorial_id, tutorial_urls, cache_dir)
        logger.debug(f"Loading notebook for tutorial '{tutorial_id}' from '{path}'")
        with open(path, "r", encoding="utf-8") as f:
            notebook = nbformat.read(f, as_version=4)
        logger.debug(f"Parsing notebook for tutorial '{tutorial_id}' to markdown")
        return notebook_to_markdown(notebook)
    except Exception as e:
        logger.error(
            f"Error getting tutorial content for tutorial_id='{tutorial_id}': {e}"
        )
        raise


async def fetch_notebook_from_github(
    tutorial_id: str, url: str, cache_dir: Path = TUTORIALS_CACHE_DIR
) -> Path:
    """
    Fetch a notebook from GitHub and cache it locally.

    Parameters
    ----------
    tutorial_id : str
        The ID of the tutorial.
    url : str
        The raw GitHub URL to the notebook file.
    cache_dir : Path, optional
        Directory to cache the notebook. Defaults to TUTORIALS_CACHE_DIR.

    Returns
    -------
    Path
        Path to the cached notebook file.

    Raises
    ------
    httpx.HTTPError
        If the download fails.

    Examples
    --------
    >>> await fetch_notebook_from_github('my_tutorial', 'https://github.com/.../my_tutorial.ipynb')
    """
    _initialize_data_dir(cache_dir)
    cache_path = _get_cached_notebook_path(tutorial_id, cache_dir)
    async with httpx.AsyncClient() as client:
        response = await client.get(url)
        response.raise_for_status()
        cache_path.write_bytes(response.content)
    logger.info(
        f"Downloaded and cached notebook for tutorial '{tutorial_id}' at '{cache_path}'"
    )
    return cache_path


def notebook_to_markdown(notebook: "nbformat.NotebookNode") -> str:
    """
    Convert a Jupyter notebook to Markdown.

    Parameters
    ----------
    notebook : nbformat.NotebookNode
        The loaded notebook object (as returned by nbformat.read).

    Returns
    -------
    str
        Markdown representation of the notebook, including code cells and outputs.

    Examples
    --------
    >>> import nbformat
    >>> with open('notebook.ipynb') as f:
    ...     nb = nbformat.read(f, as_version=4)
    >>> md = notebook_to_markdown(nb)
    >>> print(md)
    """
    markdown = []
    for cell in notebook.cells:
        if cell.cell_type == "markdown":
            markdown.append(cell.source)
        elif cell.cell_type == "code":
            markdown.append("```python")
            markdown.append(cell.source)
            markdown.append("```")
            if cell.outputs:
                markdown.append("\nOutput:")
                for output in cell.outputs:
                    if "text" in output:
                        markdown.append("```")
                        markdown.append(output["text"])
                        markdown.append("```")
                    elif "data" in output:
                        if "text/plain" in output["data"]:
                            markdown.append("```")
                            markdown.append(output["data"]["text/plain"])
                            markdown.append("```")
        markdown.append("\n---\n")
    return "\n".join(markdown)


async def _ensure_notebook_cached(
    tutorial_id: str,
    tutorial_urls: Dict[str, str] = TUTORIAL_URLS,
    cache_dir: Path = TUTORIALS_CACHE_DIR,
) -> Path:
    """
    Ensure the notebook is cached locally, fetching from GitHub if needed.

    Parameters
    ----------
    tutorial_id : str
        The ID of the tutorial.
    tutorial_urls : dict, optional
        Mapping of tutorial IDs to GitHub raw URLs. Defaults to TUTORIAL_URLS.
    cache_dir : Path, optional
        Directory to cache notebooks. Defaults to TUTORIALS_CACHE_DIR.

    Returns
    -------
    Path
        Path to the cached notebook file.

    Raises
    ------
    FileNotFoundError
        If the tutorial ID is not found in tutorial_urls.
    httpx.HTTPError
        If the download fails.

    Examples
    --------
    >>> await _ensure_notebook_cached('my_tutorial')
    """
    cache_path = _get_cached_notebook_path(tutorial_id, cache_dir)
    if not os.path.isfile(cache_path):
        url = tutorial_urls[tutorial_id]
        if not url:
            raise FileNotFoundError(
                f"No GitHub URL found for tutorial ID: {tutorial_id}"
            )
        await fetch_notebook_from_github(tutorial_id, url, cache_dir)
    return cache_path


def _get_cached_notebook_path(
    tutorial_id: str, cache_dir: Path = TUTORIALS_CACHE_DIR
) -> Path:
    """
    Get the local cache path for a tutorial notebook.

    Parameters
    ----------
    tutorial_id : str
        The ID of the tutorial.
    cache_dir : Path, optional
        Directory to cache notebooks. Defaults to TUTORIALS_CACHE_DIR.

    Returns
    -------
    Path
        Path to the cached notebook file.

    Examples
    --------
    >>> _get_cached_notebook_path('my_tutorial')
    PosixPath('.../my_tutorial.ipynb')
    """
    return Path(cache_dir) / f"{tutorial_id}.ipynb"
